import json
import time
from datetime import datetime, timezone
from enum import Enum
from fractions import Fraction
from typing import Any, Dict, List, Optional, Union

import websocket
from cachetools import Cache, LRUCache, TTLCache, func

from pycardano.address import Address
from pycardano.backend.base import (
    ALONZO_COINS_PER_UTXO_WORD,
    ChainContext,
    GenesisParameters,
    ProtocolParameters,
)
from pycardano.backend.kupo import KupoChainContextExtension, extract_asset_info
from pycardano.exception import TransactionFailedException
from pycardano.hash import DatumHash
from pycardano.network import Network
from pycardano.plutus import ExecutionUnits, PlutusV1Script, PlutusV2Script
from pycardano.serialization import RawCBOR
from pycardano.transaction import (
    Asset,
    MultiAsset,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
)
from pycardano.types import JsonDict

__all__ = ["OgmiosV5ChainContext", "KupoOgmiosV5ChainContext"]


class OgmiosQueryType(str, Enum):
    Query = "Query"
    SubmitTx = "SubmitTx"
    EvaluateTx = "EvaluateTx"


class OgmiosV5ChainContext(ChainContext):
    """Legacy Ogmios Chain Context for Ogmios v5"""

    _ws_url: str
    _network: Network
    _service_name: str
    _last_known_block_slot: int
    _last_chain_tip_fetch: float
    _genesis_param: Optional[GenesisParameters]
    _protocol_param: Optional[ProtocolParameters]
    _utxo_cache: Cache
    _datum_cache: Cache

    def __init__(
        self,
        ws_url: str,
        network: Network,
        compact_result=True,
        refetch_chain_tip_interval: Optional[float] = None,
        utxo_cache_size: int = 10000,
        datum_cache_size: int = 10000,
    ):
        self._ws_url = ws_url
        self._network = network
        self._service_name = "ogmios.v1:compact" if compact_result else "ogmios"
        self._last_known_block_slot = 0
        self._refetch_chain_tip_interval = (
            refetch_chain_tip_interval
            if refetch_chain_tip_interval is not None
            else 1000
        )
        self._last_chain_tip_fetch = 0
        self._genesis_param = None
        self._protocol_param = None
        if refetch_chain_tip_interval is None:
            self._refetch_chain_tip_interval = float(
                self.genesis_param.slot_length
                / self.genesis_param.active_slots_coefficient
            )

        self._utxo_cache = TTLCache(
            ttl=self._refetch_chain_tip_interval, maxsize=utxo_cache_size
        )
        self._datum_cache = LRUCache(maxsize=datum_cache_size)

    def _request(self, method: OgmiosQueryType, args: JsonDict) -> Any:
        ws = websocket.WebSocket()
        ws.connect(self._ws_url)
        request = json.dumps(
            {
                "type": "jsonwsp/request",
                "version": "1.0",
                "servicename": self._service_name,
                "methodname": method.value,
                "args": args,
            },
            separators=(",", ":"),
        )
        ws.send(request)
        response = ws.recv()
        ws.close()
        if "result" not in response:
            raise TransactionFailedException(
                f"Ogmios ran into an error. Reponse: {response}"
            )
        return json.loads(response)["result"]

    def _query_current_protocol_params(self) -> JsonDict:
        args = {"query": "currentProtocolParameters"}
        return self._request(OgmiosQueryType.Query, args)

    def _query_genesis_config(self) -> JsonDict:
        args = {"query": "genesisConfig"}
        return self._request(OgmiosQueryType.Query, args)

    def _query_current_epoch(self) -> int:
        args = {"query": "currentEpoch"}
        return self._request(OgmiosQueryType.Query, args)

    def _query_chain_tip(self) -> JsonDict:
        args = {"query": "chainTip"}
        return self._request(OgmiosQueryType.Query, args)

    def _query_utxos_by_address(self, address: str) -> List[List[JsonDict]]:
        args = {"query": {"utxo": [address]}}
        return self._request(OgmiosQueryType.Query, args)

    def _query_utxos_by_tx_id(self, tx_id: str, index: int) -> List[List[JsonDict]]:
        args = {"query": {"utxo": [{"txId": tx_id, "index": index}]}}
        return self._request(OgmiosQueryType.Query, args)

    def _is_chain_tip_updated(self):
        # fetch at most every twenty seconds!
        if time.time() - self._last_chain_tip_fetch < self._refetch_chain_tip_interval:
            return False
        self._last_chain_tip_fetch = time.time()
        slot = self.last_block_slot
        if self._last_known_block_slot < slot:
            self._last_known_block_slot = slot
            return True
        else:
            return False

    @staticmethod
    def _fraction_parser(fraction: str) -> Fraction:
        return Fraction(fraction)

    @property
    def protocol_param(self) -> ProtocolParameters:
        """Get current protocol parameters"""
        if not self._protocol_param or self._is_chain_tip_updated():
            self._protocol_param = self._fetch_protocol_param()
        return self._protocol_param

    def _fetch_protocol_param(self) -> ProtocolParameters:
        result = self._query_current_protocol_params()
        param = ProtocolParameters(
            min_fee_constant=result["minFeeConstant"],
            min_fee_coefficient=result["minFeeCoefficient"],
            max_block_size=result["maxBlockBodySize"],
            max_tx_size=result["maxTxSize"],
            max_block_header_size=result["maxBlockHeaderSize"],
            key_deposit=result["stakeKeyDeposit"],
            pool_deposit=result["poolDeposit"],
            pool_influence=self._fraction_parser(result["poolInfluence"]),
            monetary_expansion=self._fraction_parser(result["monetaryExpansion"]),
            treasury_expansion=self._fraction_parser(result["treasuryExpansion"]),
            decentralization_param=self._fraction_parser(
                result.get("decentralizationParameter", "0/1")
            ),
            extra_entropy=result.get("extraEntropy", ""),
            protocol_major_version=result["protocolVersion"]["major"],
            protocol_minor_version=result["protocolVersion"]["minor"],
            min_utxo=self._get_min_utxo(),
            min_pool_cost=result["minPoolCost"],
            price_mem=self._fraction_parser(result["prices"]["memory"]),
            price_step=self._fraction_parser(result["prices"]["steps"]),
            max_tx_ex_mem=result["maxExecutionUnitsPerTransaction"]["memory"],
            max_tx_ex_steps=result["maxExecutionUnitsPerTransaction"]["steps"],
            max_block_ex_mem=result["maxExecutionUnitsPerBlock"]["memory"],
            max_block_ex_steps=result["maxExecutionUnitsPerBlock"]["steps"],
            max_val_size=result["maxValueSize"],
            collateral_percent=result["collateralPercentage"],
            max_collateral_inputs=result["maxCollateralInputs"],
            coins_per_utxo_word=result.get(
                "coinsPerUtxoWord", ALONZO_COINS_PER_UTXO_WORD
            ),
            coins_per_utxo_byte=result.get("coinsPerUtxoByte", 0),
            cost_models=self._parse_cost_models(result),
        )

        return param

    def _get_min_utxo(self) -> int:
        result = self._query_genesis_config()
        return result["protocolParameters"]["minUtxoValue"]

    def _parse_cost_models(self, ogmios_result: JsonDict) -> Dict[str, Dict[str, int]]:
        ogmios_cost_models = ogmios_result.get("costModels", {})

        cost_models = {}
        if "plutus:v1" in ogmios_cost_models:
            cost_models["PlutusV1"] = ogmios_cost_models["plutus:v1"].copy()
        if "plutus:v2" in ogmios_cost_models:
            cost_models["PlutusV2"] = ogmios_cost_models["plutus:v2"].copy()
        return cost_models

    @property
    def genesis_param(self) -> GenesisParameters:
        """Get chain genesis parameters"""

        @func.lru_cache(maxsize=10)
        def _genesis_param_cache(slot) -> GenesisParameters:
            return self._fetch_genesis_param()

        return _genesis_param_cache(self.last_block_slot)

    def _fetch_genesis_param(self) -> GenesisParameters:
        result = self._query_genesis_config()
        start_str = result["systemStart"].split(".")[0]
        dt = datetime.fromisoformat(start_str.replace("Z", "+00:00"))
        dt = dt.replace(tzinfo=timezone.utc)
        system_start_unix = int(dt.timestamp())
        return GenesisParameters(
            active_slots_coefficient=self._fraction_parser(
                result["activeSlotsCoefficient"]
            ),
            update_quorum=result["updateQuorum"],
            max_lovelace_supply=result["maxLovelaceSupply"],
            network_magic=result["networkMagic"],
            epoch_length=result["epochLength"],
            system_start=system_start_unix,
            slots_per_kes_period=result["slotsPerKesPeriod"],
            slot_length=result["slotLength"],
            max_kes_evolutions=result["maxKesEvolutions"],
            security_param=result["securityParameter"],
        )

    @property
    def network(self) -> Network:
        """Get current network"""
        return self._network

    @property
    def epoch(self) -> int:
        """Current epoch number"""
        return self._query_current_epoch()

    @property
    @func.ttl_cache(ttl=1)
    def last_block_slot(self) -> int:
        result = self._query_chain_tip()
        slot = result["slot"]
        return slot

    def _utxos(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        key = (self.last_block_slot, address)
        if key in self._utxo_cache:
            return self._utxo_cache[key]

        utxos = self._utxos_ogmios(address)

        self._utxo_cache[key] = utxos

        return utxos

    def _check_utxo_unspent(self, tx_id: str, index: int) -> bool:
        """Check whether an UTxO is unspent with Ogmios.

        Args:
            tx_id (str): transaction id.
            index (int): transaction index.
        """
        results = self._query_utxos_by_tx_id(tx_id, index)
        return len(results) > 0

    def _utxos_ogmios(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address with Ogmios.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        results = self._query_utxos_by_address(address)

        utxos = []
        for result in results:
            utxos.append(self._utxo_from_ogmios_result(result))

        return utxos

    def utxo_by_tx_id(self, tx_id: str, index: int) -> Optional[UTxO]:
        """Get a UTxO associated with a tx_id and index.

        Args:
            tx_id (str): The transaction id.
            index (int): The index for the UTxO at the specified transaction.

        Returns:
            Optional[UTxO]: Return a UTxO if exists at the tx_id and index.
        """
        results = self._query_utxos_by_tx_id(tx_id, index)
        assert len(results) < 2, f"Query for UTxO {tx_id}#{index} should be unique!"

        utxos = []
        for result in results:
            utxos.append(self._utxo_from_ogmios_result(result))

        if len(utxos) > 0:
            return utxos[0]
        return None

    def _utxo_from_ogmios_result(self, result) -> UTxO:
        in_ref = result[0]
        output = result[1]
        tx_in = TransactionInput.from_primitive([in_ref["txId"], in_ref["index"]])
        lovelace_amount = output["value"]["coins"]
        script = output.get("script", None)
        if script:
            if "plutus:v2" in script:
                script = PlutusV2Script(bytes.fromhex(script["plutus:v2"]))
            elif "plutus:v1" in script:
                script = PlutusV1Script(bytes.fromhex(script["plutus:v1"]))
            else:
                raise ValueError("Unknown plutus script type")
        datum_hash = (
            DatumHash.from_primitive(output["datumHash"])
            if output.get("datumHash", None)
            else None
        )
        datum = None
        if output["datum"] and output["datum"] != output["datumHash"]:
            datum = RawCBOR(bytes.fromhex(output["datum"]))
        if not output["value"]["assets"]:
            tx_out = TransactionOutput(
                Address.from_primitive(output["address"]),
                amount=lovelace_amount,
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )
        else:
            multi_assets = MultiAsset()

            for asset, quantity in output["value"]["assets"].items():
                policy_hex, policy, asset_name_hex = extract_asset_info(asset)
                multi_assets.setdefault(policy, Asset())[asset_name_hex] = quantity

            tx_out = TransactionOutput(
                Address.from_primitive(output["address"]),
                amount=Value(lovelace_amount, multi_assets),
                datum_hash=datum_hash,
                datum=datum,
                script=script,
            )
        utxo = UTxO(tx_in, tx_out)
        return utxo

    def submit_tx_cbor(self, cbor: Union[bytes, str]):
        """Submit a transaction to the blockchain.

        Args:
            cbor (Union[bytes, str]): The transaction to be submitted.

        Raises:
            :class:`InvalidArgumentException`: When the transaction is invalid.
            :class:`TransactionFailedException`: When fails to submit the transaction to blockchain.
        """
        if isinstance(cbor, bytes):
            cbor = cbor.hex()

        args = {"submit": cbor}
        result = self._request(OgmiosQueryType.SubmitTx, args)
        if "SubmitFail" in result:
            raise TransactionFailedException(result["SubmitFail"])

    def evaluate_tx_cbor(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
        """Evaluate execution units of a transaction.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be evaluated.

        Returns:
            Dict[str, ExecutionUnits]: A list of execution units calculated for each of the transaction's redeemers

        Raises:
            :class:`TransactionFailedException`: When fails to evaluate the transaction.
        """
        if isinstance(cbor, bytes):
            cbor = cbor.hex()

        args = {"evaluate": cbor}
        result = self._request(OgmiosQueryType.EvaluateTx, args)
        if "EvaluationResult" not in result:
            raise TransactionFailedException(result)
        else:
            for k in result["EvaluationResult"].keys():
                result["EvaluationResult"][k] = ExecutionUnits(
                    result["EvaluationResult"][k]["memory"],
                    result["EvaluationResult"][k]["steps"],
                )
            return result["EvaluationResult"]


def KupoOgmiosV5ChainContext(
    ws_url: str,
    network: Network,
    compact_result=True,
    refetch_chain_tip_interval: Optional[float] = None,
    utxo_cache_size: int = 10000,
    datum_cache_size: int = 10000,
    kupo_url: Optional[str] = None,
) -> KupoChainContextExtension:
    return KupoChainContextExtension(
        OgmiosV5ChainContext(
            ws_url,
            network,
            compact_result,
            refetch_chain_tip_interval,
            utxo_cache_size,
            datum_cache_size,
        ),
        kupo_url,
    )



class PyCardanoException(Exception):
    pass


class DecodingException(PyCardanoException):
    pass


class InvalidKeyTypeException(PyCardanoException):
    pass


class InvalidAddressInputException(PyCardanoException):
    pass


class InvalidDataException(PyCardanoException):
    pass


class InvalidArgumentException(PyCardanoException):
    pass


class InvalidOperationException(PyCardanoException):
    pass


class SerializeException(PyCardanoException):
    pass


class DeserializeException(PyCardanoException):
    pass


class InvalidTransactionException(PyCardanoException):
    pass


class TransactionBuilderException(PyCardanoException):
    pass


class TransactionFailedException(PyCardanoException):
    pass


class UTxOSelectionException(PyCardanoException):
    pass


class InsufficientUTxOBalanceException(UTxOSelectionException):
    pass


class MaxInputCountExceededException(UTxOSelectionException):
    pass


class InputUTxODepletedException(UTxOSelectionException):
    pass


class CardanoCliError(PyCardanoException):
    pass



"""
BIP32 implemented on curve ED25519
Paper: https://github.com/LedgerHQ/orakolo/blob/master/papers/Ed25519_BIP%20Final.pdf
This is a modified version of https://github.com/johnoliverdriscoll/py-edhd/blob/master/src/edhd/__init__.py

"""

from __future__ import annotations

import hashlib
import hmac
import unicodedata
from binascii import hexlify, unhexlify
from typing import Optional, Tuple

from mnemonic import Mnemonic
from nacl import bindings

from pycardano.logging import logger

__all__ = ["BIP32ED25519PrivateKey", "BIP32ED25519PublicKey", "HDWallet"]


SUPPORTED_MNEMONIC_LANGS = {
    "english",
    "french",
    "italian",
    "japanese",
    "chinese_simplified",
    "chinese_traditional",
    "korean",
    "spanish",
}


class BIP32ED25519PrivateKey:
    def __init__(self, private_key: bytes, chain_code: bytes):
        self.private_key = private_key
        self.left = self.private_key[:32]
        self.right = self.private_key[32:]
        self.chain_code = chain_code
        self.public_key = bindings.crypto_scalarmult_ed25519_base_noclamp(self.left)

    def sign(self, message: bytes) -> bytes:
        r = bindings.crypto_core_ed25519_scalar_reduce(
            hashlib.sha512(self.right + message).digest(),
        )
        R = bindings.crypto_scalarmult_ed25519_base_noclamp(r)
        hram = bindings.crypto_core_ed25519_scalar_reduce(
            hashlib.sha512(R + self.public_key + message).digest(),
        )
        S = bindings.crypto_core_ed25519_scalar_add(
            bindings.crypto_core_ed25519_scalar_mul(hram, self.left),
            r,
        )
        return R + S


class BIP32ED25519PublicKey:
    def __init__(self, public_key: bytes, chain_code: bytes):
        self.public_key = public_key
        self.chain_code = chain_code

    @classmethod
    def from_private_key(
        cls, private_key: BIP32ED25519PrivateKey
    ) -> BIP32ED25519PublicKey:
        return cls(private_key.public_key, private_key.chain_code)

    def verify(self, signature, message):
        return bindings.crypto_sign_open(signature + message, self.public_key)


def _Fk(message, secret):
    return hmac.new(secret, message, hashlib.sha512).digest()


class HDWallet:
    """
    Hierarchical Deterministic Wallet for Cardano
    """

    def __init__(
        self,
        root_xprivate_key: Optional[bytes] = None,
        root_public_key: Optional[bytes] = None,
        root_chain_code: Optional[bytes] = None,
        xprivate_key: Optional[bytes] = None,
        public_key: Optional[bytes] = None,
        chain_code: Optional[bytes] = None,
        path: str = "m",
        seed: Optional[bytes] = None,
        mnemonic: Optional[str] = None,
        passphrase: Optional[str] = None,
        entropy: Optional[str] = None,
    ):
        self._root_xprivate_key = root_xprivate_key
        self._root_public_key = root_public_key
        self._root_chain_code = root_chain_code

        self._xprivate_key = xprivate_key
        self._public_key = public_key
        self._chain_code = chain_code

        self._path = path

        self._seed = seed
        self._mnemonic = mnemonic
        self._passphrase = passphrase
        self._entropy = entropy

    @classmethod
    def from_seed(
        cls,
        seed: str,
        entropy: Optional[str] = None,
        passphrase: Optional[str] = None,
        mnemonic: Optional[str] = None,
    ) -> HDWallet:
        """
        Create an HDWallet instance from master key.

        Args:
            seed: Master key of 96 bytes from seed hex string.
            entropy: Entropy hex string, default to ``None``.
            passphrase: Mnemonic passphrase or password, default to ``None``.
            mnemonic: Mnemonic words, default to ``None``.

        Returns:
            HDWallet -- Hierarchical Deterministic Wallet instance.
        """

        seed_converted = bytearray(bytes.fromhex(seed))
        seed_modified = cls._tweak_bits(seed_converted)

        kL, c = seed_modified[:32], seed_modified[64:]

        # root public key
        A = bindings.crypto_scalarmult_ed25519_base_noclamp(kL)

        return cls(
            root_xprivate_key=seed_modified[:64],
            root_public_key=A,
            root_chain_code=c,
            xprivate_key=seed_modified[:64],
            public_key=A,
            chain_code=c,
            seed=seed_modified,
            mnemonic=mnemonic,
            passphrase=passphrase,
            entropy=entropy,
        )

    @classmethod
    def from_mnemonic(cls, mnemonic: str, passphrase: str = "") -> HDWallet:
        """
        Create master key and HDWallet from Mnemonic words.

        Args:
            mnemonic: Mnemonic words.
            passphrase: Mnemonic passphrase or password, default to ``None``.

        Returns:
            HDWallet -- Hierarchical Deterministic Wallet instance.
        """

        if not cls.is_mnemonic(mnemonic=mnemonic):
            raise ValueError("Invalid mnemonic words.")

        mnemonic = unicodedata.normalize("NFKD", mnemonic)
        entropy = Mnemonic(language="english").to_entropy(words=mnemonic)
        seed = cls._generate_seed(passphrase, entropy)

        return cls.from_seed(
            seed=hexlify(seed).decode(),
            mnemonic=mnemonic,
            entropy=hexlify(entropy).decode("utf-8"),
            passphrase=passphrase,
        )

    @classmethod
    def from_entropy(cls, entropy: str, passphrase: str = "") -> HDWallet:
        """
        Create master key and HDWallet from Mnemonic words.

        Args:
            entropy: Entropy hex string.
            passphrase: Mnemonic passphrase or password, default to ``None``.

        Returns:
            HDWallet -- Hierarchical Deterministic Wallet instance.
        """

        if not cls.is_entropy(entropy):
            raise ValueError("Invalid entropy")

        seed = cls._generate_seed(passphrase, bytearray.fromhex(entropy))
        return cls.from_seed(seed=hexlify(seed).decode(), entropy=entropy)

    @classmethod
    def _generate_seed(cls, passphrase: str, entropy: bytearray) -> bytearray:
        return bytearray(
            hashlib.pbkdf2_hmac(
                "sha512",
                password=passphrase.encode(),
                salt=entropy,
                iterations=4096,
                dklen=96,
            )
        )

    @classmethod
    def _tweak_bits(cls, seed: bytearray) -> bytes:
        """
        Modify seed based on Icarus master node derivation scheme.

        The process follows
        `CIP-0003#Wallet Key Generation <https://github.com/cardano-foundation/CIPs/tree/master/CIP-0003>`_.

        Process:
            - clear the lowest 3 bits
            - clear the highest 3 bits
            - set the highest second bit

        Args:
            seed: Seed in bytearray

        Returns:
            modified seed in bytes.
        """
        seed[0] &= 0b11111000
        seed[31] &= 0b00011111
        seed[31] |= 0b01000000

        return bytes(seed)

    def _copy_hdwallet(self):
        """
        Create a new instance of HDWallet
        """

        return HDWallet(
            root_xprivate_key=self._root_xprivate_key,
            root_public_key=self._root_public_key,
            root_chain_code=self._root_chain_code,
            xprivate_key=self._xprivate_key,
            public_key=self._public_key,
            chain_code=self._chain_code,
            path=self._path,
            seed=self._seed,
            mnemonic=self._mnemonic,
            passphrase=self._passphrase,
            entropy=self._entropy,
        )

    def derive_from_path(self, path: str, private: bool = True) -> HDWallet:
        """
        Derive keys from a path following CIP-1852 specifications.

        Args:
            path: Derivation path for the key generation.
            private: whether to derive private child keys or public child keys.

        Returns:
            HDWallet instance with keys derived

        Examples:
            >>> mnemonic_words = "test walk nut penalty hip pave soap entry language right filter choice"
            >>> hdwallet = HDWallet.from_mnemonic(mnemonic_words)
            >>> child_hdwallet = hdwallet.derive_from_path("m/1852'/1815'/0'/0/0")
            >>> child_hdwallet.public_key.hex()
            '73fea80d424276ad0978d4fe5310e8bc2d485f5f6bb3bf87612989f112ad5a7d'
        """

        if path[:2] != "m/":
            raise ValueError(
                'Bad path, please insert like this type of path "m/0\'/0"! '
            )

        derived_hdwallet = self._copy_hdwallet()
        for index in path.lstrip("m/").split("/"):
            if index.endswith("'"):
                derived_hdwallet = derived_hdwallet.derive(
                    int(index[:-1]), private=private, hardened=True
                )
            else:
                derived_hdwallet = derived_hdwallet.derive(
                    int(index), private=private, hardened=False
                )

        return derived_hdwallet

    def derive(
        self,
        index: int,
        private: bool = True,
        hardened: bool = False,
    ) -> HDWallet:
        """
        Returns a new HDWallet derived from given index.

        Args:
            index: Derivation index.
            private: whether to derive private child keys or public child keys.
            hardened: whether to derive hardened address. Default to False.

        Returns:
            HDWallet instance with keys derived

        Examples:
            >>> mnemonic_words = "test walk nut penalty hip pave soap entry language right filter choice"
            >>> hdwallet = HDWallet.from_mnemonic(mnemonic_words)
            >>> hdwallet = hdwallet.derive(1852, hardened=True)
            >>> hdwallet = hdwallet.derive(1815, hardened=True)
            >>> hdwallet = hdwallet.derive(0, hardened=True)
            >>> hdwallet = hdwallet.derive(0)
            >>> hdwallet = hdwallet.derive(0)
            >>> hdwallet.public_key.hex()
            '73fea80d424276ad0978d4fe5310e8bc2d485f5f6bb3bf87612989f112ad5a7d'
        """

        if not isinstance(index, int):
            raise ValueError("Bad index, Please import only integer number!")

        if hardened:
            index += 2**31

        if private:
            if self._xprivate_key is None:
                raise ValueError("Missing private key. Can't do private derivation.")

            private_node = (
                self._xprivate_key[:32],
                self._xprivate_key[32:],
                self._public_key,
                self._chain_code,
                self._path,
            )

            if any(x is None for x in private_node):
                raise ValueError(f"None values in private node: {private_node}")

            return self._derive_private_child_key_by_index(private_node, index)  # type: ignore

        public_node = (
            self._public_key,
            self._chain_code,
            self._path,
        )

        if any(x is None for x in public_node):
            raise ValueError(f"None values in public node: {public_node}")

        return self._derive_public_child_key_by_index(public_node, index)  # type: ignore

    def _derive_private_child_key_by_index(
        self, private_pnode: Tuple[bytes, bytes, bytes, bytes, str], index: int
    ) -> HDWallet:
        """
        Derive private child keys from parent node.

        PROCESS:
          1. encode i 4-bytes little endian, il = encode_U32LE(i)
          2. if i is less than 2^31
               - compute Z   = HMAC-SHA512(key=c, Data=0x02 | A | il )
               - compute c_  = HMAC-SHA512(key=c, Data=0x03 | A | il )
             else
               - compute Z   = HMAC-SHA512(key=c, Data=0x00 | kL | kR | il )
               - compute c_  = HMAC-SHA512(key=c, Data=0x01 | kL | kR | il )
          3. ci = lowest_32bytes(c_)
          4. set ZL = highest_28bytes(Z)
             set ZR = lowest_32bytes(Z)
          5. compute kL_i:
                zl_  = LEBytes_to_int(ZL)
                kL_  = LEBytes_to_int(kL)
                kLi_ = zl_*8 + kL_
                if kLi_ % order == 0: child does not exist
                kL_i = int_to_LEBytes(kLi_)
          6. compute kR_i
                zr_  = LEBytes_to_int(ZR)
                kR_  = LEBytes_to_int(kR)
                kRi_ = (zr_ + kRn_) % 2^256
                kR_i = int_to_LEBytes(kRi_)
          7. compute A
                A = kLi_.G
          8. return ((kL_i,kR_i), A_i, c)

        Args:
            private_pnode: ((kLP,kRP), AP, cP). (kLP,kRP) is 64 bytes parent private eddsa key,
                AP is 32 btyes parent public key, cP is 32 btyes parent chain code.
            index: child index to compute (hardened if >= 0x80000000)

        Returns:
            HDWallet with child node derived.

        """
        # unpack argument
        (kLP, kRP, AP, cP, path) = private_pnode
        assert 0 <= index < 2**32

        i_bytes = index.to_bytes(4, "little")

        # compute Z,c
        if index < 2**31:
            # regular child
            Z = _Fk(b"\x02" + AP + i_bytes, cP)
            c = _Fk(b"\x03" + AP + i_bytes, cP)[32:]
        else:
            # harderned child
            Z = _Fk(b"\x00" + (kLP + kRP) + i_bytes, cP)
            c = _Fk(b"\x01" + (kLP + kRP) + i_bytes, cP)[32:]

        ZL, ZR = Z[:28], Z[32:]

        # compute KLi
        kLn = int.from_bytes(ZL, "little") * 8 + int.from_bytes(kLP, "little")

        # compute KRi
        kRn = (int.from_bytes(ZR, "little") + int.from_bytes(kRP, "little")) % 2**256

        kL = kLn.to_bytes(32, "little")
        kR = kRn.to_bytes(32, "little")

        # compue Ai
        A = bindings.crypto_scalarmult_ed25519_base_noclamp(kL)

        # compute path
        path += "/" + str(index)

        derived_hdwallet = HDWallet(
            root_xprivate_key=self._root_xprivate_key,
            root_public_key=self._root_public_key,
            root_chain_code=self._root_chain_code,
            xprivate_key=kL + kR,
            public_key=A,
            chain_code=c,
            path=path,
            seed=self._seed,
            mnemonic=self._mnemonic,
            passphrase=self._passphrase,
            entropy=self._entropy,
        )

        return derived_hdwallet

    def _derive_public_child_key_by_index(
        self, public_pnode: Tuple[bytes, bytes, str], index: int
    ) -> HDWallet:
        """
        Derive public child keys from parent node.

        Args:
            public_pnode: (AP, cP). AP is 32 btyes parent public key, cP is 32 btyes parent chain code.
            index: child index to compute (hardened if >= 0x80000000)

        Returns:
            HDWallet with child node derived.
        """
        # unpack argument
        (AP, cP, path) = public_pnode
        assert 0 <= index < 2**32

        i_bytes = index.to_bytes(4, "little")

        # compute Z,c
        if index < 2**31:
            # regular child
            Z = _Fk(b"\x02" + AP + i_bytes, cP)
            c = _Fk(b"\x03" + AP + i_bytes, cP)[32:]
        else:
            # can't derive hardened child from public keys
            raise ValueError("Cannot derive hardened index with public key")

        ZL = Z[:28]

        # compute ZLi
        ZLint = int.from_bytes(ZL, "little")

        # compue Ai
        A = bindings.crypto_core_ed25519_add(
            AP,
            bindings.crypto_scalarmult_ed25519_base_noclamp(
                (8 * ZLint).to_bytes(32, "little")
            ),
        )

        # compute path
        path += "/" + str(index)

        derived_hdwallet = HDWallet(
            root_xprivate_key=self._root_xprivate_key,
            root_public_key=self._root_public_key,
            root_chain_code=self._root_chain_code,
            xprivate_key=self._xprivate_key,
            public_key=A,
            chain_code=c,
            path=path,
            seed=self._seed,
            mnemonic=self._mnemonic,
            passphrase=self._passphrase,
            entropy=self._entropy,
        )

        return derived_hdwallet

    @property
    def root_xprivate_key(self):
        return self._root_xprivate_key

    @property
    def root_public_key(self):
        return self._root_public_key

    @property
    def root_chain_code(self):
        return self._root_chain_code

    @property
    def xprivate_key(self):
        return self._xprivate_key

    @property
    def public_key(self):
        return self._public_key

    @property
    def chain_code(self):
        return self._chain_code

    @staticmethod
    def generate_mnemonic(language: str = "english", strength: int = 256) -> str:
        """
        Generate mnemonic words.

        Args:
            language (str): language for the mnemonic words.
            strength (int): length of the mnemoic words. Valid values are 128/160/192/224/256.

        Returns:
            mnemonic (str): mnemonic words.
        """

        if language and language not in SUPPORTED_MNEMONIC_LANGS:
            raise ValueError(
                "invalid language, use only this options english, french, "
                "italian, spanish, chinese_simplified, chinese_traditional, japanese or korean languages."
            )
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError(
                "Strength should be one of the following "
                "[128, 160, 192, 224, 256], but it is not (%d)." % strength
            )

        return Mnemonic(language=language).generate(strength=strength)

    @staticmethod
    def is_mnemonic(mnemonic: str, language: Optional[str] = None) -> bool:
        """
        Check if mnemonic words are valid.

        Args:
            mnemonic (str): Mnemonic words in string format.
            language (Optional[str]): Mnemonic language, default to None.

        Returns:
            bool. Whether the input mnemonic words is valid.
        """

        if language and language not in SUPPORTED_MNEMONIC_LANGS:
            raise ValueError(
                "invalid language, use only this options english, french, "
                "italian, spanish, chinese_simplified, chinese_traditional, japanese or korean languages."
            )
        try:
            mnemonic = unicodedata.normalize("NFKD", mnemonic)
            if language:
                return Mnemonic(language=language).check(mnemonic=mnemonic)

            for _language in SUPPORTED_MNEMONIC_LANGS:
                if Mnemonic(language=_language).check(mnemonic=mnemonic) is True:
                    return True
        except ValueError:
            logger.warning(
                "The input mnemonic words are not valid. Words should be in string format seperated by space."
            )
        return False

    @staticmethod
    def is_entropy(entropy: str) -> bool:
        """
        Check entropy hex string.

        Args:
            entropy: entropy converted from mnemonic words.

        Returns:
            bool. Whether entropy is valid or not.
        """

        try:
            return len(unhexlify(entropy)) in [16, 20, 24, 28, 32]
        except ValueError:
            logger.warning("The input entropy is not valid.")
            return False



"""
Conditional cbor2 import module.

This module provides a centralized location for importing cbor2,
with support for both the C extension (cbor2) and pure Python (cbor2pure) versions.
Set the environment variable CBOR_C_EXTENSION=1 to use the C extension.
"""

import os

if os.getenv("CBOR_C_EXTENSION", "0") == "1":
    import cbor2  # noqa: F401
else:
    import cbor2pure as cbor2  # type: ignore  # noqa: F401



"""Defines interfaces for client codes to interact (read/write) with the blockchain."""

from dataclasses import dataclass
from fractions import Fraction
from typing import Dict, List, Optional, Union

from pycardano.address import Address
from pycardano.exception import InvalidArgumentException
from pycardano.logging import log_state
from pycardano.network import Network
from pycardano.plutus import ExecutionUnits
from pycardano.transaction import Transaction, UTxO
from pycardano.types import typechecked

__all__ = [
    "GenesisParameters",
    "ProtocolParameters",
    "ChainContext",
    "ALONZO_COINS_PER_UTXO_WORD",
]

ALONZO_COINS_PER_UTXO_WORD = 34482


@dataclass(frozen=True)
class GenesisParameters:
    """Cardano genesis parameters"""

    active_slots_coefficient: Fraction

    update_quorum: int

    max_lovelace_supply: int

    network_magic: int

    epoch_length: int

    system_start: int

    slots_per_kes_period: int

    slot_length: int

    max_kes_evolutions: int

    security_param: int


@dataclass(frozen=True)
class ProtocolParameters:
    """Cardano protocol parameters"""

    min_fee_constant: int

    min_fee_coefficient: int

    max_block_size: int

    max_tx_size: int

    max_block_header_size: int

    key_deposit: int

    pool_deposit: int

    pool_influence: Fraction

    monetary_expansion: Fraction

    treasury_expansion: Fraction

    decentralization_param: Fraction

    extra_entropy: str

    protocol_major_version: int

    protocol_minor_version: int

    min_utxo: int

    min_pool_cost: int

    price_mem: Fraction

    price_step: Fraction

    max_tx_ex_mem: int

    max_tx_ex_steps: int

    max_block_ex_mem: int

    max_block_ex_steps: int

    max_val_size: int

    collateral_percent: int

    max_collateral_inputs: int

    coins_per_utxo_word: int

    coins_per_utxo_byte: int

    cost_models: Dict[str, Dict[str, int]]
    """A dict contains cost models for Plutus. The key will be "PlutusV1", "PlutusV2", etc.
    The value will be a dict of cost model parameters."""

    maximum_reference_scripts_size: Optional[Dict[str, int]] = None

    min_fee_reference_scripts: Optional[Dict[str, float]] = None


@typechecked
class ChainContext:
    """Interfaces through which the library interacts with Cardano blockchain."""

    @property
    def protocol_param(self) -> ProtocolParameters:
        """Get current protocol parameters"""
        raise NotImplementedError()

    @property
    def genesis_param(self) -> GenesisParameters:
        """Get chain genesis parameters"""
        raise NotImplementedError()

    @property
    def network(self) -> Network:
        """Cet current network"""
        raise NotImplementedError()

    @property
    def epoch(self) -> int:
        """Current epoch number"""
        raise NotImplementedError()

    @property
    def last_block_slot(self) -> int:
        """Slot number of last block"""
        raise NotImplementedError()

    def utxos(self, address: Union[str, Address]) -> List[UTxO]:
        """Get all UTxOs associated with an address.

        Args:
            address (Union[str, Address]): An address, potentially bech32 encoded

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        return self._utxos(str(address))

    def _utxos(self, address: str) -> List[UTxO]:
        """Get all UTxOs associated with an address.

        Args:
            address (str): An address encoded with bech32.

        Returns:
            List[UTxO]: A list of UTxOs.
        """
        raise NotImplementedError()

    @log_state
    def submit_tx(self, tx: Union[Transaction, bytes, str]):
        """Submit a transaction to the blockchain.

        Args:
            tx (Union[Transaction, bytes, str]): The transaction to be submitted.

        Raises:
            :class:`InvalidArgumentException`: When the transaction is invalid.
            :class:`TransactionFailedException`: When fails to submit the transaction to blockchain.
        """
        if isinstance(tx, Transaction):
            return self.submit_tx_cbor(tx.to_cbor())
        elif isinstance(tx, bytes) or isinstance(tx, str):
            return self.submit_tx_cbor(tx)
        else:
            raise InvalidArgumentException(
                f"Invalid transaction type: {type(tx)}, expected Transaction, bytes, or str"
            )

    def submit_tx_cbor(self, cbor: Union[bytes, str]):
        """Submit a transaction to the blockchain.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be submitted.

        Raises:
            :class:`InvalidArgumentException`: When the transaction is invalid.
            :class:`TransactionFailedException`: When fails to submit the transaction to blockchain.
        """
        raise NotImplementedError()

    def evaluate_tx(self, tx: Transaction) -> Dict[str, ExecutionUnits]:
        """Evaluate execution units of a transaction.

        Args:
            transaction (Transaction): The transaction to be evaluated.

        Returns:
            List[ExecutionUnits]: A list of execution units calculated for each of the transaction's redeemers
        """
        return self.evaluate_tx_cbor(tx.to_cbor())

    def evaluate_tx_cbor(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
        """Evaluate execution units of a transaction.

        Args:
            cbor (Union[bytes, str]): The serialized transaction to be evaluated.

        Returns:
            List[ExecutionUnits]: A list of execution units calculated for each of the transaction's redeemers
        """
        raise NotImplementedError()



"""Definitions of transaction-related data types."""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Callable, List, Optional, Type, Union

from cbor2 import CBORTag
from nacl.encoding import RawEncoder
from nacl.hash import blake2b
from pprintpp import pformat

from pycardano.address import Address
from pycardano.cbor import cbor2
from pycardano.certificate import Certificate
from pycardano.exception import InvalidDataException
from pycardano.governance import ProposalProcedure, VotingProcedures
from pycardano.hash import (
    TRANSACTION_HASH_SIZE,
    AuxiliaryDataHash,
    ConstrainedBytes,
    DatumHash,
    ScriptDataHash,
    ScriptHash,
    TransactionId,
    VerificationKeyHash,
)
from pycardano.metadata import AuxiliaryData
from pycardano.nativescript import NativeScript
from pycardano.network import Network
from pycardano.plutus import Datum, PlutusScript, RawPlutusData
from pycardano.serialization import (
    ArrayCBORSerializable,
    CBORSerializable,
    DictBase,
    DictCBORSerializable,
    MapCBORSerializable,
    NonEmptyOrderedSet,
    OrderedSet,
    Primitive,
    default_encoder,
    limit_primitive_type,
    list_hook,
)
from pycardano.types import typechecked
from pycardano.witness import TransactionWitnessSet

__all__ = [
    "TransactionInput",
    "AssetName",
    "Asset",
    "MultiAsset",
    "Value",
    "TransactionOutput",
    "UTxO",
    "TransactionBody",
    "Transaction",
    "Withdrawals",
]

_MAX_INT64 = (1 << 63) - 1
_MIN_INT64 = -(1 << 63)


@dataclass(repr=False)
class TransactionInput(ArrayCBORSerializable):
    transaction_id: TransactionId

    index: int

    def __hash__(self):
        return hash(str(self.transaction_id) + str(self.index))


class AssetName(ConstrainedBytes):
    MAX_SIZE = 32

    def __repr__(self):
        return f"AssetName({self.payload})"


@typechecked
class Asset(DictCBORSerializable):
    KEY_TYPE = AssetName

    VALUE_TYPE = int

    def normalize(self) -> Asset:
        """Normalize the Asset by removing zero values."""
        for k, v in list(self.items()):
            if v == 0:
                self.pop(k)

        return self

    def union(self, other: Asset) -> Asset:
        return self + other

    def __add__(self, other: Asset) -> Asset:
        new_asset = deepcopy(self)
        for n in other:
            new_asset[n] = new_asset.get(n, 0) + other[n]
        return new_asset.normalize()

    def __iadd__(self, other: Asset) -> Asset:
        new_item = self + other
        self.data = new_item.data  # type: ignore[has-type]
        return self.normalize()

    def __sub__(self, other: Asset) -> Asset:
        new_asset = deepcopy(self)
        for n in other:
            new_asset[n] = new_asset.get(n, 0) - other[n]
        return new_asset.normalize()

    def __eq__(self, other):
        if not isinstance(other, Asset):
            return False
        else:
            if len(self) != len(other):
                return False
            for n in self:
                if n not in other or self[n] != other[n]:
                    return False
            return True

    def __le__(self, other: Asset) -> bool:
        for n in self:
            if n not in other or self[n] > other[n]:
                return False
        return True

    def __lt__(self, other: Asset):
        return self <= other and self != other

    def __ge__(self, other: Asset) -> bool:
        for n in other:
            if n not in self or self[n] < other[n]:
                return False
        return True

    def __gt__(self, other: Asset) -> bool:
        return self >= other and self != other

    @classmethod
    @limit_primitive_type(dict)
    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
        res = super().from_primitive(value)
        # pop zero values
        for n, v in list(res.items()):
            if v == 0:
                res.pop(n)
        return res

    def to_shallow_primitive(self) -> dict:
        x = deepcopy(self).normalize()
        return super(self.__class__, x).to_shallow_primitive()


@typechecked
class MultiAsset(DictCBORSerializable):
    KEY_TYPE = ScriptHash

    VALUE_TYPE = Asset

    def union(self, other: MultiAsset) -> MultiAsset:
        return self + other

    def normalize(self) -> MultiAsset:
        """Normalize the MultiAsset by removing zero values."""
        for k, v in list(self.items()):
            v.normalize()
            if len(v) == 0:
                self.pop(k)
        return self

    def __add__(self, other):
        new_multi_asset = deepcopy(self)
        for p in other:
            new_multi_asset[p] = new_multi_asset.get(p, Asset()) + other[p]
        return new_multi_asset.normalize()

    def __iadd__(self, other):
        new_item = self + other
        self.data = new_item.data
        return self.normalize()

    def __sub__(self, other: MultiAsset) -> MultiAsset:
        new_multi_asset = deepcopy(self)
        for p in other:
            new_multi_asset[p] = new_multi_asset.get(p, Asset()) - other[p]
        return new_multi_asset.normalize()

    def __eq__(self, other):
        if not isinstance(other, MultiAsset):
            return False
        else:
            if len(self) != len(other):
                return False
            for p in self:
                if p not in other or self[p] != other[p]:
                    return False
            return True

    def __ge__(self, other: MultiAsset) -> bool:
        for n in other:
            if n not in self:
                return False
            if not self[n] >= other[n]:
                return False
        return True

    def __gt__(self, other: MultiAsset) -> bool:
        return self >= other and self != other

    def __le__(self, other: MultiAsset):
        for p in self:
            if p not in other:
                return False
            if not self[p] <= other[p]:
                return False
        return True

    def __lt__(self, other: MultiAsset):
        return self <= other and self != other

    def filter(
        self, criteria=Callable[[ScriptHash, AssetName, int], bool]
    ) -> MultiAsset:
        """Filter items by criteria.

        Args:
            criteria: A function that takes in three input arguments (policy_id, asset_name, amount) and returns a
                bool. If returned value is True, then the asset will be kept, otherwise discarded.

        Returns:
            A new filtered MultiAsset object.
        """
        new_multi_asset = MultiAsset()

        for p in self:
            for n in self[p]:
                if criteria(p, n, self[p][n]):
                    if p not in new_multi_asset:
                        new_multi_asset[p] = Asset()
                    new_multi_asset[p][n] = self[p][n]

        return new_multi_asset

    def count(self, criteria=Callable[[ScriptHash, AssetName, int], bool]) -> int:
        """Count number of distinct assets that satisfy a certain criteria.

        Args:
            criteria: A function that takes in three input arguments (policy_id, asset_name, amount) and returns a
                bool.

        Returns:
            int: Total number of distinct assets that satisfy the criteria.
        """
        count = 0
        for p in self:
            for n in self[p]:
                if criteria(p, n, self[p][n]):
                    count += 1

        return count

    @classmethod
    @limit_primitive_type(dict)
    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
        res = super().from_primitive(value)
        # pop empty values
        for n, v in list(res.items()):
            if not v:
                res.pop(n)
        return res

    def to_shallow_primitive(self) -> dict:
        x = deepcopy(self).normalize()
        return super(self.__class__, x).to_shallow_primitive()


@typechecked
@dataclass(repr=False)
class Value(ArrayCBORSerializable):
    coin: int = 0
    """Amount of ADA"""

    multi_asset: MultiAsset = field(default_factory=MultiAsset)
    """Multi-assets associated with the UTxO"""

    def union(self, other: Union[Value, int]) -> Value:
        return self + other

    def __add__(self, other: Union[Value, int]):
        if isinstance(other, int):
            other = Value(other)
        return Value(self.coin + other.coin, self.multi_asset + other.multi_asset)

    def __iadd__(self, other: Union[Value, int]):
        new_item = self + other
        self.coin = new_item.coin
        self.multi_asset = new_item.multi_asset
        return self

    def __sub__(self, other: Union[Value, int]) -> Value:
        if isinstance(other, int):
            other = Value(other)
        return Value(self.coin - other.coin, self.multi_asset - other.multi_asset)

    def __eq__(self, other):
        if not isinstance(other, (Value, int)):
            return False
        else:
            if isinstance(other, int):
                other = Value(other)
            return self.coin == other.coin and self.multi_asset == other.multi_asset

    def __le__(self, other: Union[Value, int]):
        if isinstance(other, int):
            other = Value(other)
        return self.coin <= other.coin and self.multi_asset <= other.multi_asset

    def __lt__(self, other: Union[Value, int]):
        return self <= other and self != other

    def __ge__(self, other: Union[Value, int]):
        if isinstance(other, int):
            other = Value(other)
        return self.coin >= other.coin and self.multi_asset >= other.multi_asset

    def __gt__(self, other: Union[Value, int]):
        return self >= other and self != other

    def to_shallow_primitive(self):
        if self.multi_asset:
            return super().to_shallow_primitive()
        else:
            return self.coin


@dataclass(repr=False)
class _Script(ArrayCBORSerializable):
    _TYPE: int = field(init=False, default=0)

    script: Union[NativeScript, PlutusScript]

    def __post_init__(self):
        if isinstance(self.script, NativeScript):
            self._TYPE = 0
        elif isinstance(self.script, PlutusScript):
            self._TYPE = self.script.version

    @classmethod
    def from_primitive(
        cls: Type[_Script], values: List[Primitive], type_args: Optional[tuple] = None
    ) -> _Script:
        if values[0] == 0:
            return cls(NativeScript.from_primitive(values[1]))
        assert isinstance(values[1], bytes)
        assert isinstance(values[0], int)
        return cls(PlutusScript.from_version(values[0], values[1]))


@dataclass(repr=False)
class _DatumOption(ArrayCBORSerializable):
    _TYPE: int = field(init=False, default=0)

    datum: Union[DatumHash, Any]

    def __post_init__(self):
        if isinstance(self.datum, DatumHash):
            self._TYPE = 0
        else:
            self._TYPE = 1

    def to_shallow_primitive(self) -> Primitive:
        data: Union[CBORTag, DatumHash]
        if self._TYPE == 1:
            data = CBORTag(24, cbor2.dumps(self.datum, default=default_encoder))
        else:
            data = self.datum
        return [self._TYPE, data]

    @classmethod
    def from_primitive(
        cls: Type[_DatumOption],
        values: List[Primitive],
        type_args: Optional[tuple] = None,
    ) -> _DatumOption:
        if values[0] == 0:
            assert isinstance(values[1], bytes)
            return _DatumOption(DatumHash(values[1]))
        else:
            assert isinstance(values[1], CBORTag)
            v = cbor2.loads(values[1].value)
            if isinstance(v, CBORTag):
                return _DatumOption(RawPlutusData.from_primitive(v))
            else:
                return _DatumOption(v)


@dataclass(repr=False)
class _ScriptRef(CBORSerializable):
    script: _Script

    def to_primitive(self) -> Primitive:
        return CBORTag(24, cbor2.dumps(self.script, default=default_encoder))

    @classmethod
    def from_primitive(
        cls: Type[_ScriptRef], value: List[Primitive], type_args: Optional[tuple] = None
    ) -> _ScriptRef:
        assert isinstance(value, CBORTag)
        return cls(_Script.from_primitive(cbor2.loads(value.value)))


@dataclass(repr=False)
class _TransactionOutputPostAlonzo(MapCBORSerializable):
    address: Address = field(metadata={"key": 0})

    amount: Union[int, Value] = field(metadata={"key": 1})

    datum: Optional[_DatumOption] = field(
        default=None, metadata={"key": 2, "optional": True}
    )

    script_ref: Optional[_ScriptRef] = field(
        default=None, metadata={"key": 3, "optional": True}
    )

    @property
    def script(
        self,
    ) -> Optional[Union[NativeScript, PlutusScript]]:
        if self.script_ref:
            return self.script_ref.script.script
        else:
            return None


@dataclass(repr=False)
class _TransactionOutputLegacy(ArrayCBORSerializable):
    address: Address

    amount: Union[int, Value]

    datum_hash: Optional[DatumHash] = field(default=None, metadata={"optional": True})


@dataclass(repr=False)
class TransactionOutput(CBORSerializable):
    address: Address

    amount: Union[Value]

    datum_hash: Optional[DatumHash] = None

    datum: Optional[Datum] = None

    script: Optional[Union[NativeScript, PlutusScript]] = None

    post_alonzo: Optional[bool] = False

    def __post_init__(self):
        if isinstance(self.address, str):
            self.address = Address.from_primitive(self.address)
        if isinstance(self.amount, int):
            self.amount = Value(self.amount)

    def validate(self):
        super().validate()
        if isinstance(self.amount, Value) and (
            self.amount.coin < 0
            or self.amount.multi_asset.count(lambda p, n, v: v < 0) > 0
        ):
            raise InvalidDataException(
                f"Transaction output cannot have negative amount of ADA or "
                f"native asset: \n {self.amount}"
            )

    @property
    def lovelace(self) -> int:
        return self.amount.coin

    def to_primitive(self) -> Primitive:
        if self.datum or self.script or self.post_alonzo:
            datum = (
                _DatumOption(self.datum_hash or self.datum)
                if self.datum is not None or self.datum_hash is not None
                else None
            )
            script_ref = (
                _ScriptRef(_Script(self.script)) if self.script is not None else None
            )
            return _TransactionOutputPostAlonzo(
                self.address, self.amount, datum, script_ref
            ).to_primitive()
        else:
            return _TransactionOutputLegacy(
                self.address, self.amount, self.datum_hash
            ).to_primitive()

    @classmethod
    def from_primitive(
        cls: Type[TransactionOutput],
        value: List[Primitive],
        type_args: Optional[tuple] = None,
    ) -> TransactionOutput:
        if isinstance(value, list):
            output = _TransactionOutputLegacy.from_primitive(value)
            return cls(output.address, output.amount, datum_hash=output.datum_hash)
        else:
            output = _TransactionOutputPostAlonzo.from_primitive(value)
            datum = output.datum.datum if output.datum else None
            if isinstance(datum, DatumHash):
                return cls(
                    output.address,
                    output.amount,
                    datum_hash=datum,
                    script=output.script,
                )
            else:
                return cls(
                    output.address,
                    output.amount,
                    datum=datum,
                    script=output.script,
                )


@dataclass(repr=False)
class UTxO(ArrayCBORSerializable):
    input: TransactionInput

    output: TransactionOutput

    def __repr__(self):
        return pformat(vars(self))

    def __hash__(self):
        return hash(blake2b(self.input.to_cbor() + self.output.to_cbor(), 32))


class Withdrawals(DictCBORSerializable):
    """A disctionary of reward addresses to reward withdrawal amount.

    Key is address bytes, value is an integer.

    Examples:

        >>> address = Address.from_primitive("stake_test1upyz3gk6mw5he20apnwfn96cn9rscgvmmsxc9r86dh0k66gswf59n")
        >>> Withdrawals({bytes(address): 1000000}) # doctest: +NORMALIZE_WHITESPACE
        {b'\\xe0H(\\xa2\\xda\\xdb\\xa9|\\xa9\\xfd\\x0c\\xdc\\x99\\x97X\\x99G\\x0c!\\x9b\\xdc\\r\\x82\\x8c\\xfam\\xdfmi':
        1000000}
    """

    KEY_TYPE = bytes

    VALUE_TYPE = int


@dataclass(repr=False)
class TransactionBody(MapCBORSerializable):
    inputs: Union[List[TransactionInput], OrderedSet[TransactionInput]] = field(
        default_factory=OrderedSet,
        metadata={"key": 0},
    )

    outputs: List[TransactionOutput] = field(
        default_factory=list,
        metadata={"key": 1, "object_hook": list_hook(TransactionOutput)},
    )

    fee: int = field(default=0, metadata={"key": 2})

    ttl: Optional[int] = field(default=None, metadata={"key": 3, "optional": True})

    certificates: Optional[
        Union[List[Certificate], NonEmptyOrderedSet[Certificate]]
    ] = field(
        default=None,
        metadata={
            "key": 4,
            "optional": True,
        },
    )

    withdraws: Optional[Withdrawals] = field(
        default=None, metadata={"key": 5, "optional": True}
    )

    update: Any = field(default=None, metadata={"key": 6, "optional": True})

    auxiliary_data_hash: Optional[AuxiliaryDataHash] = field(
        default=None, metadata={"key": 7, "optional": True}
    )

    validity_start: Optional[int] = field(
        default=None, metadata={"key": 8, "optional": True}
    )

    mint: Optional[MultiAsset] = field(
        default=None, metadata={"key": 9, "optional": True}
    )

    script_data_hash: Optional[ScriptDataHash] = field(
        default=None, metadata={"key": 11, "optional": True}
    )

    collateral: Optional[
        Union[List[TransactionInput], NonEmptyOrderedSet[TransactionInput]]
    ] = field(
        default=None,
        metadata={
            "key": 13,
            "optional": True,
        },
    )

    required_signers: Optional[
        Union[List[VerificationKeyHash], NonEmptyOrderedSet[VerificationKeyHash]]
    ] = field(
        default=None,
        metadata={
            "key": 14,
            "optional": True,
        },
    )

    network_id: Optional[Network] = field(
        default=None, metadata={"key": 15, "optional": True}
    )

    collateral_return: Optional[TransactionOutput] = field(
        default=None, metadata={"key": 16, "optional": True}
    )

    total_collateral: Optional[int] = field(
        default=None, metadata={"key": 17, "optional": True}
    )

    reference_inputs: Optional[
        Union[List[TransactionInput], NonEmptyOrderedSet[TransactionInput]]
    ] = field(
        default=None,
        metadata={
            "key": 18,
            "optional": True,
        },
    )

    voting_procedures: Optional[VotingProcedures] = field(
        default=None, metadata={"key": 19, "optional": True}
    )

    proposal_procedures: Optional[NonEmptyOrderedSet[ProposalProcedure]] = field(
        default=None, metadata={"key": 20, "optional": True}
    )

    current_treasury_value: Optional[int] = field(
        default=None, metadata={"key": 21, "optional": True}
    )

    donation: Optional[int] = field(
        default=None, metadata={"key": 22, "optional": True}
    )

    def validate(self):
        if (
            self.mint
            and self.mint.count(lambda p, n, v: v < _MIN_INT64 or v > _MAX_INT64) > 0
        ):
            raise InvalidDataException(
                f"Mint amount must be between {_MIN_INT64} and {_MAX_INT64}. \n Mint amount: {self.mint}"
            )

    def hash(self) -> bytes:
        return blake2b(self.to_cbor(), TRANSACTION_HASH_SIZE, encoder=RawEncoder)  # type: ignore

    @property
    def id(self) -> TransactionId:
        return TransactionId(self.hash())


@dataclass(repr=False)
class Transaction(ArrayCBORSerializable):
    transaction_body: TransactionBody

    transaction_witness_set: TransactionWitnessSet

    valid: Optional[bool] = field(default=True, metadata={"optional": True})

    auxiliary_data: Optional[AuxiliaryData] = None

    @property
    def json_type(self) -> str:
        return (
            "Unwitnessed Tx ConwayEra"
            if self.transaction_witness_set.vkey_witnesses is None
            else "Signed Tx ConwayEra"
        )

    @property
    def json_description(self) -> str:
        return "Ledger Cddl Format"

    @property
    def id(self) -> TransactionId:
        return self.transaction_body.id



import os
from functools import partial
from typing import Any, Dict

import typeguard

# https://github.com/python/typing/issues/182#issuecomment-199532520
JsonDict = Dict[str, Any]


def typechecked(func=None, *args, **kwargs):
    if os.getenv("PYCARDANO_NO_TYPE_CHECK", "False").lower() in ("true", "1"):
        if func is None:
            return partial(typechecked, *args, **kwargs)
        return func
    return typeguard.typechecked(func, *args, **kwargs)


def check_type(*args, **kwargs):
    if os.getenv("PYCARDANO_NO_TYPE_CHECK", "False").lower() in ("true", "1"):
        return None
    return typeguard.check_type(*args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, ClassVar, List, Optional, Type, Union

from cbor2 import CBORTag
from nacl.encoding import RawEncoder
from nacl.hash import blake2b

from pycardano.exception import DeserializeException, InvalidArgumentException
from pycardano.hash import AUXILIARY_DATA_HASH_SIZE, AuxiliaryDataHash
from pycardano.nativescript import NativeScript
from pycardano.plutus import PlutusV1Script, PlutusV2Script, PlutusV3Script
from pycardano.serialization import (
    ArrayCBORSerializable,
    CBORSerializable,
    DictCBORSerializable,
    MapCBORSerializable,
    Primitive,
    limit_primitive_type,
    list_hook,
)

__all__ = ["Metadata", "ShelleyMarryMetadata", "AlonzoMetadata", "AuxiliaryData"]


class Metadata(DictCBORSerializable):
    KEY_TYPE = int  # transaction_metadatum_label, see https://github.com/cardano-foundation/CIPs/tree/master/CIP-0010
    VALUE_TYPE = Any

    MAX_ITEM_SIZE = 64
    INTERNAL_TYPES = (dict, list, int, bytes, str)

    def _validate(self):
        def _validate_type_and_size(data):
            if not isinstance(data, self.INTERNAL_TYPES):
                raise InvalidArgumentException(
                    f"A value in Metadata has to be one of {self.INTERNAL_TYPES}, "
                    f"got {type(data)} instead."
                )
            if isinstance(data, bytes):
                if len(data) > self.MAX_ITEM_SIZE:
                    raise InvalidArgumentException(
                        f"The size of {data} exceeds {self.MAX_ITEM_SIZE} bytes."
                        "Use pycardano.serialization.ByteString for long bytes."
                    )
            elif isinstance(data, str):
                if len(data.encode("utf-8")) > self.MAX_ITEM_SIZE:
                    raise InvalidArgumentException(
                        f"The size of {data} exceeds {self.MAX_ITEM_SIZE} bytes."
                    )
            elif isinstance(data, list):
                for item in data:
                    _validate_type_and_size(item)
            elif isinstance(data, dict):
                for key in data:
                    _validate_type_and_size(data[key])

        for k in self:
            if not isinstance(k, self.KEY_TYPE):
                raise InvalidArgumentException(
                    f"Keys in the first layer of Metadata has to be {self.KEY_TYPE}, "
                    f"got {type(k)} instead."
                )
            _validate_type_and_size(self[k])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validate()


@dataclass
class ShelleyMarryMetadata(ArrayCBORSerializable):
    metadata: Metadata
    native_scripts: Optional[List[NativeScript]] = field(
        default=None, metadata={"object_hook": list_hook(NativeScript)}
    )


@dataclass
class AlonzoMetadata(MapCBORSerializable):
    TAG: ClassVar[int] = 259

    metadata: Optional[Metadata] = field(
        default=None, metadata={"optional": True, "key": 0}
    )
    native_scripts: Optional[List[NativeScript]] = field(
        default=None,
        metadata={"optional": True, "key": 1, "object_hook": list_hook(NativeScript)},
    )
    plutus_v1_scripts: Optional[List[PlutusV1Script]] = field(
        default=None, metadata={"optional": True, "key": 2}
    )
    plutus_v2_scripts: Optional[List[PlutusV2Script]] = field(
        default=None, metadata={"optional": True, "key": 3}
    )
    plutus_v3_scripts: Optional[List[PlutusV3Script]] = field(
        default=None, metadata={"optional": True, "key": 4}
    )

    def to_primitive(self) -> Primitive:
        return CBORTag(AlonzoMetadata.TAG, super(AlonzoMetadata, self).to_primitive())

    @classmethod
    @limit_primitive_type(CBORTag)
    def from_primitive(cls: Type[AlonzoMetadata], value: CBORTag) -> AlonzoMetadata:
        if not hasattr(value, "tag"):
            raise DeserializeException(
                f"{value} does not match the data schema of AlonzoMetadata."
            )
        elif value.tag != cls.TAG:
            raise DeserializeException(
                f"Expect CBOR tag: {cls.TAG}, got {value.tag} instead."
            )
        return super(AlonzoMetadata, cls).from_primitive(value.value)


@dataclass
class AuxiliaryData(CBORSerializable):
    data: Union[Metadata, ShelleyMarryMetadata, AlonzoMetadata]

    def to_primitive(self) -> Primitive:
        return self.data.to_primitive()

    @classmethod
    def from_primitive(
        cls: Type[AuxiliaryData], value: Primitive, type_args: Optional[tuple] = None
    ) -> AuxiliaryData:
        for t in [AlonzoMetadata, ShelleyMarryMetadata, Metadata]:
            # The schema of metadata in different eras are mutually exclusive, so we can try deserializing
            # them one by one without worrying about mismatch.
            try:
                return AuxiliaryData(t.from_primitive(value))  # type: ignore
            except DeserializeException:
                pass
        raise DeserializeException(f"Couldn't parse auxiliary data: {value}")

    def hash(self) -> AuxiliaryDataHash:
        return AuxiliaryDataHash(
            blake2b(self.to_cbor(), AUXILIARY_DATA_HASH_SIZE, encoder=RawEncoder)  # type: ignore
        )



# flake8: noqa
# Copyright (c) 2017, 2020 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Reference implementation for Bech32/Bech32m and segwit addresses."""


from enum import Enum


class Encoding(Enum):
    """Enumeration type to list the various supported encodings."""

    BECH32 = 1
    BECH32M = 2


CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"
BECH32M_CONST = 0x2BC830A3


def bech32_polymod(values):
    """Internal function that computes the Bech32 checksum."""
    generator = [0x3B6A57B2, 0x26508E6D, 0x1EA119FA, 0x3D4233DD, 0x2A1462B3]
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1FFFFFF) << 5 ^ value
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrp_expand(hrp):
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


def bech32_verify_checksum(hrp, data):
    """Verify a checksum given HRP and converted data characters."""
    const = bech32_polymod(bech32_hrp_expand(hrp) + data)
    if const == 1:
        return Encoding.BECH32
    if const == BECH32M_CONST:
        return Encoding.BECH32M
    return None


def bech32_create_checksum(hrp, data, spec):
    """Compute the checksum values given HRP and data."""
    values = bech32_hrp_expand(hrp) + data
    const = BECH32M_CONST if spec == Encoding.BECH32M else 1
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ const
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def bech32_encode(hrp, data, spec):
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data, spec)
    return hrp + "1" + "".join([CHARSET[d] for d in combined])


def bech32_decode(bech):
    """Validate a Bech32/Bech32m string, and determine HRP and data."""
    if (any(ord(x) < 33 or ord(x) > 126 for x in bech)) or (
        bech.lower() != bech and bech.upper() != bech
    ):
        return (None, None, None)
    bech = bech.lower()
    pos = bech.rfind("1")
    if pos < 1 or pos + 7 > len(bech) or len(bech) > 108:
        return (None, None, None)
    if not all(x in CHARSET for x in bech[pos + 1 :]):
        return (None, None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos + 1 :]]
    spec = bech32_verify_checksum(hrp, data)
    if spec is None:
        return (None, None, None)
    return (hrp, data[:-6], spec)


def convertbits(data, frombits, tobits, pad=True):
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def decode(addr):
    """Decode a segwit address."""
    _, data, _ = bech32_decode(addr)
    decoded = convertbits(data, 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 108:
        return None
    return decoded


def encode(hrp, witprog):
    """Encode a segwit address."""
    ret = bech32_encode(hrp, convertbits(witprog, 8, 5), 0)
    if bech32_decode(ret) == (None, None, None):
        return None
    return ret



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field, fields
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from pycardano import RedeemerMap
from pycardano.address import Address, AddressType
from pycardano.backend.base import ChainContext
from pycardano.certificate import (
    Certificate,
    PoolRegistration,
    PoolRetirement,
    RegDRepCert,
    StakeAndVoteDelegation,
    StakeCredential,
    StakeDelegation,
    StakeDeregistration,
    StakeDeregistrationConway,
    StakeRegistration,
    StakeRegistrationAndDelegation,
    StakeRegistrationAndDelegationAndVoteDelegation,
    StakeRegistrationAndVoteDelegation,
    StakeRegistrationConway,
    VoteDelegation,
)
from pycardano.coinselection import (
    LargestFirstSelector,
    RandomImproveMultiAsset,
    UTxOSelector,
)
from pycardano.exception import (
    InsufficientUTxOBalanceException,
    InvalidArgumentException,
    InvalidTransactionException,
    TransactionBuilderException,
    UTxOSelectionException,
)
from pycardano.governance import (
    Anchor,
    GovAction,
    GovActionId,
    GovActionIdToVotingProcedure,
    ProposalProcedure,
    Vote,
    Voter,
    VotingProcedure,
    VotingProcedures,
)
from pycardano.hash import DatumHash, ScriptDataHash, ScriptHash, VerificationKeyHash
from pycardano.key import ExtendedSigningKey, SigningKey, VerificationKey
from pycardano.logging import log_state, logger
from pycardano.metadata import AuxiliaryData
from pycardano.nativescript import NativeScript, ScriptAll, ScriptAny, ScriptPubkey
from pycardano.plutus import (
    CostModels,
    Datum,
    ExecutionUnits,
    PlutusScript,
    PlutusV1Script,
    PlutusV2Script,
    PlutusV3Script,
    Redeemer,
    RedeemerKey,
    Redeemers,
    RedeemerTag,
    RedeemerValue,
    ScriptType,
    datum_hash,
    script_hash,
)
from pycardano.serialization import NonEmptyOrderedSet, OrderedSet
from pycardano.transaction import (
    Asset,
    AssetName,
    MultiAsset,
    Transaction,
    TransactionBody,
    TransactionInput,
    TransactionOutput,
    UTxO,
    Value,
    Withdrawals,
)
from pycardano.utils import fee, max_tx_fee, min_lovelace_post_alonzo, script_data_hash
from pycardano.witness import TransactionWitnessSet, VerificationKeyWitness

__all__ = ["TransactionBuilder"]

FAKE_VKEY = VerificationKey.from_primitive(
    bytes.fromhex("5797dc2cc919dfec0bb849551ebdf30d96e5cbe0f33f734a87fe826db30f7ef9")
)

# Ed25519 signature of a 32-bytes message (TX hash) will have length of 64
FAKE_TX_SIGNATURE = bytes.fromhex(
    "577ccb5b487b64e396b0976c6f71558e52e44ad254db7d06dfb79843e5441a5d763dd42a"
    "dcf5e8805d70373722ebbce62a58e3f30dd4560b9a898b8ceeab6a03"
)


@dataclass
class TransactionBuilder:
    """A class builder that makes it easy to build a transaction."""

    context: ChainContext

    utxo_selectors: List[UTxOSelector] = field(
        default_factory=lambda: [LargestFirstSelector(), RandomImproveMultiAsset()]
    )

    execution_memory_buffer: float = 0.2
    """Additional amount of execution memory (in ratio) that will be on top of estimation"""

    execution_step_buffer: float = 0.2
    """Additional amount of execution step (in ratio) that will be added on top of estimation"""

    fee_buffer: Optional[int] = field(default=None)
    """Additional amount of fee (in lovelace) that will be added on top of estimation."""

    ttl: Optional[int] = field(default=None)

    validity_start: Optional[int] = field(default=None)

    auxiliary_data: Optional[AuxiliaryData] = field(default=None)

    native_scripts: Optional[List[NativeScript]] = field(default=None)

    mint: Optional[MultiAsset] = field(default=None)

    required_signers: Optional[List[VerificationKeyHash]] = field(default=None)

    collaterals: NonEmptyOrderedSet[UTxO] = field(
        default_factory=lambda: NonEmptyOrderedSet[UTxO]()
    )

    certificates: Optional[List[Certificate]] = field(default=None)

    withdrawals: Optional[Withdrawals] = field(default=None)

    reference_inputs: Set[Union[UTxO, TransactionInput]] = field(
        init=False, default_factory=lambda: set()
    )

    witness_override: Optional[int] = field(default=None)

    initial_stake_pool_registration: Optional[bool] = field(default=False)

    use_redeemer_map: Optional[bool] = field(default=True)
    """Whether to serialize redeemers as a map or a list. Default is True."""

    voting_procedures: Optional[VotingProcedures] = field(init=False, default=None)

    proposal_procedures: Optional[NonEmptyOrderedSet[ProposalProcedure]] = field(
        init=False, default=None
    )

    current_treasury_value: Optional[int] = field(init=False, default=None)

    donation: Optional[int] = field(init=False, default=None)

    _inputs: List[UTxO] = field(init=False, default_factory=lambda: [])

    _potential_inputs: List[UTxO] = field(init=False, default_factory=lambda: [])

    _excluded_inputs: List[UTxO] = field(init=False, default_factory=lambda: [])

    _input_addresses: List[Union[Address, str]] = field(
        init=False, default_factory=lambda: []
    )

    _outputs: List[TransactionOutput] = field(init=False, default_factory=lambda: [])

    _fee: int = field(init=False, default=0)

    _datums: Dict[DatumHash, Datum] = field(init=False, default_factory=lambda: {})

    _collateral_return: Optional[TransactionOutput] = field(init=False, default=None)

    collateral_return_threshold: int = 1_000_000
    """The minimum amount of lovelace above which
    the remaining collateral (total_collateral_amount - actually_used_amount) will be returned."""

    _total_collateral: Optional[int] = field(init=False, default=None)

    _inputs_to_redeemers: Dict[UTxO, Redeemer] = field(
        init=False, default_factory=lambda: {}
    )

    _minting_script_to_redeemers: List[Tuple[ScriptType, Optional[Redeemer]]] = field(
        init=False, default_factory=lambda: []
    )

    _withdrawal_script_to_redeemers: List[Tuple[ScriptType, Optional[Redeemer]]] = (
        field(init=False, default_factory=lambda: [])
    )

    _certificate_script_to_redeemers: List[Tuple[ScriptType, Optional[Redeemer]]] = (
        field(init=False, default_factory=lambda: [])
    )

    _inputs_to_scripts: Dict[UTxO, ScriptType] = field(
        init=False, default_factory=lambda: {}
    )

    _reference_scripts: List[Union[NativeScript, PlutusScript]] = field(
        init=False, default_factory=lambda: []
    )

    _should_estimate_execution_units: Optional[bool] = field(init=False, default=None)

    def add_input(self, utxo: UTxO) -> TransactionBuilder:
        """Add a specific UTxO to transaction's inputs.

        Args:
            utxo (UTxO): UTxO to be added.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        self.inputs.append(utxo)
        if utxo.output.script:
            self._reference_scripts.append(utxo.output.script)
        return self

    def _consolidate_redeemer(self, redeemer):
        if self._should_estimate_execution_units is None:
            if redeemer.ex_units:
                self._should_estimate_execution_units = False
            else:
                self._should_estimate_execution_units = True
                redeemer.ex_units = ExecutionUnits(0, 0)
        else:
            if not self._should_estimate_execution_units and not redeemer.ex_units:
                raise InvalidArgumentException(
                    f"All redeemers need to provide execution units if the firstly "
                    f"added redeemer specifies execution units. \n"
                    f"Added redeemers: {self._redeemer_list} \n"
                    f"New redeemer: {redeemer}"
                )
            if self._should_estimate_execution_units:
                if redeemer.ex_units:
                    raise InvalidArgumentException(
                        f"No redeemer should provide execution units if the firstly "
                        f"added redeemer didn't provide execution units. \n"
                        f"Added redeemers: {self._redeemer_list} \n"
                        f"New redeemer: {redeemer}"
                    )
                else:
                    redeemer.ex_units = ExecutionUnits(0, 0)

    def add_script_input(
        self,
        utxo: UTxO,
        script: Optional[Union[UTxO, NativeScript, PlutusScript]] = None,
        datum: Optional[Datum] = None,
        redeemer: Optional[Redeemer] = None,
    ) -> TransactionBuilder:
        """Add a script UTxO to transaction's inputs.

        Args:
            utxo (UTxO): Script UTxO to be added.
            script (Optional[Union[UTxO, NativeScript, PlutusScript]]):
                A plutus script.
                If not provided, the script will be inferred from the input UTxO (first arg of this method).
                The script can also be a specific UTxO whose output contains an inline script.
            datum (Optional[Datum]): A plutus datum to unlock the UTxO.
            redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        if not utxo.output.address.address_type.name.startswith("SCRIPT"):
            raise InvalidArgumentException(
                f"Expect the output address of utxo to be script type, "
                f"but got {utxo.output.address.address_type} instead."
            )

        if (
            utxo.output.datum_hash
            and datum is not None
            and utxo.output.datum_hash != datum_hash(datum)
        ):
            raise InvalidArgumentException(
                f"Datum hash in transaction output is {utxo.output.datum_hash}, "
                f"but actual datum hash from input datum is {datum_hash(datum)}."
            )
        if (
            datum is not None
            and utxo.output.datum_hash is None
            and utxo.output.datum is not None
        ):
            raise InvalidArgumentException(
                f"Inline Datum found in transaction output {utxo.input}, "
                "so attaching a Datum to the transaction input manually is not allowed."
            )

        if datum is not None:
            self.datums[datum_hash(datum)] = datum

        if redeemer:
            if redeemer.tag is not None and redeemer.tag != RedeemerTag.SPEND:
                raise InvalidArgumentException(
                    f"Expect the redeemer tag's type to be {RedeemerTag.SPEND}, "
                    f"but got {redeemer.tag} instead."
                )
            redeemer.tag = RedeemerTag.SPEND
            self._consolidate_redeemer(redeemer)
            self._inputs_to_redeemers[utxo] = redeemer

        input_script_hash = utxo.output.address.payment_part

        # collect potential scripts to fulfill the input
        candidate_scripts: List[
            Tuple[
                Union[NativeScript, PlutusScript],
                Optional[UTxO],
            ]
        ] = []
        if utxo.output.script:
            candidate_scripts.append((utxo.output.script, utxo))
        elif not script:
            for i in self.context.utxos(utxo.output.address):
                if i.output.script:
                    candidate_scripts.append((i.output.script, i))
        elif isinstance(script, UTxO):
            if script.output.script is None:
                raise InvalidArgumentException(
                    f"Expect the output of the reference UTxO {utxo}"
                    " to have a script, but got None instead."
                )
            candidate_scripts.append((script.output.script, script))
        else:
            candidate_scripts.append((script, None))

        found_valid_script = False
        for candidate_script, candidate_utxo in candidate_scripts:
            if script_hash(candidate_script) != input_script_hash:
                continue

            found_valid_script = True
            self._inputs_to_scripts[utxo] = candidate_script

            if candidate_utxo is not None and candidate_utxo != utxo:
                self.reference_inputs.add(candidate_utxo)
                self._reference_scripts.append(candidate_script)
            break
        if not found_valid_script:
            raise InvalidArgumentException(
                f"Cannot find a valid script to fulfill the input UTxO: {utxo.input}."
                "Supplied scripts do not match the payment part of the input address."
            )

        self.inputs.append(utxo)
        return self

    def add_minting_script(
        self,
        script: Union[UTxO, NativeScript, PlutusScript],
        redeemer: Optional[Redeemer] = None,
    ) -> TransactionBuilder:
        """Add a minting script along with its datum and redeemer to this transaction.

        Args:
            script (Union[UTxO, PlutusScript): A plutus script.
            redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        if redeemer:
            if redeemer.tag is not None and redeemer.tag != RedeemerTag.MINT:
                raise InvalidArgumentException(
                    f"Expect the redeemer tag's type to be {RedeemerTag.MINT}, "
                    f"but got {redeemer.tag} instead."
                )
            redeemer.tag = RedeemerTag.MINT
            self._consolidate_redeemer(redeemer)

        if isinstance(script, UTxO):
            assert script.output.script is not None
            self._minting_script_to_redeemers.append((script.output.script, redeemer))
            self.reference_inputs.add(script)
            self._reference_scripts.append(script.output.script)
        else:
            self._minting_script_to_redeemers.append((script, redeemer))
        return self

    def add_withdrawal_script(
        self,
        script: Union[UTxO, NativeScript, PlutusScript],
        redeemer: Optional[Redeemer] = None,
    ) -> TransactionBuilder:
        """Add a withdrawal script along with its redeemer to this transaction.

        Args:
            script (Union[UTxO, PlutusScript]): A plutus script.
            redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        if redeemer:
            if redeemer.tag is not None and redeemer.tag != RedeemerTag.WITHDRAWAL:
                raise InvalidArgumentException(
                    f"Expect the redeemer tag's type to be {RedeemerTag.WITHDRAWAL}, "
                    f"but got {redeemer.tag} instead."
                )
            redeemer.tag = RedeemerTag.WITHDRAWAL
            self._consolidate_redeemer(redeemer)

        if isinstance(script, UTxO):
            assert script.output.script is not None
            self._withdrawal_script_to_redeemers.append(
                (script.output.script, redeemer)
            )
            self.reference_inputs.add(script)
            self._reference_scripts.append(script.output.script)
        else:
            self._withdrawal_script_to_redeemers.append((script, redeemer))
        return self

    def add_certificate_script(
        self,
        script: Union[UTxO, NativeScript, PlutusScript],
        redeemer: Optional[Redeemer] = None,
    ) -> TransactionBuilder:
        """Add a certificate script along with its redeemer to this transaction.
        WARNING: The order of operations matters.
        The index of the redeemer will be set to the index of the last certificate added.

        Args:
            script (Union[UTxO, PlutusScript]): A plutus script.
            redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        if redeemer:
            if redeemer.tag is not None and redeemer.tag != RedeemerTag.CERTIFICATE:
                raise InvalidArgumentException(
                    f"Expect the redeemer tag's type to be {RedeemerTag.CERTIFICATE}, "
                    f"but got {redeemer.tag} instead."
                )
            assert self.certificates is not None and len(self.certificates) >= 1, (
                "self.certificates is None. redeemer.index needs to be set to the index of the corresponding"
                "certificate (defaulting to the last certificate) however no certificates could be found"
            )
            redeemer.index = len(self.certificates) - 1
            redeemer.tag = RedeemerTag.CERTIFICATE
            self._consolidate_redeemer(redeemer)

        if isinstance(script, UTxO):
            assert script.output.script is not None
            self._certificate_script_to_redeemers.append(
                (script.output.script, redeemer)
            )
            self.reference_inputs.add(script)
            self._reference_scripts.append(script.output.script)
        else:
            self._certificate_script_to_redeemers.append((script, redeemer))
        return self

    def add_input_address(self, address: Union[Address, str]) -> TransactionBuilder:
        """Add an address to transaction's input address.
        Unlike :meth:`add_input`, which deterministically adds a UTxO to the transaction's inputs, `add_input_address`
        will not immediately select any UTxO when called. Instead, it will delegate UTxO selection to
        :class:`UTxOSelector`s of the builder when :meth:`build` is called.

        Args:
            address (Union[Address, str]): Address to be added.

        Returns:
            TransactionBuilder: The current transaction builder.
        """
        self.input_addresses.append(address)
        return self

    def add_output(
        self,
        tx_out: TransactionOutput,
        datum: Optional[Datum] = None,
        add_datum_to_witness: bool = False,
    ) -> TransactionBuilder:
        """Add a transaction output.

        Args:
            tx_out (TransactionOutput): The transaction output to be added.
            datum (Datum): Attach a datum hash to this transaction output.
            add_datum_to_witness (bool): Optionally add the actual datum to transaction witness set. Defaults to False.

        Returns:
            TransactionBuilder: Current transaction builder.
        """
        if datum is not None:
            tx_out.datum_hash = datum_hash(datum)
        self.outputs.append(tx_out)
        if datum is not None and add_datum_to_witness:
            self.datums[datum_hash(datum)] = datum
        return self

    @property
    def inputs(self) -> List[UTxO]:
        return self._inputs

    @property
    def potential_inputs(self) -> List[UTxO]:
        return self._potential_inputs

    @property
    def excluded_inputs(self) -> List[UTxO]:
        return self._excluded_inputs

    @excluded_inputs.setter
    def excluded_inputs(self, excluded_inputs: List[UTxO]):
        self._excluded_inputs = excluded_inputs

    @property
    def input_addresses(self) -> List[Union[Address, str]]:
        return self._input_addresses

    @property
    def outputs(self) -> List[TransactionOutput]:
        return self._outputs

    @property
    def fee(self) -> int:
        return self._fee

    @fee.setter
    def fee(self, fee: int):
        self._fee = fee

    @property
    def all_scripts(self) -> List[ScriptType]:
        scripts: Dict[ScriptHash, ScriptType] = {}
        s: ScriptType

        if self.native_scripts:
            for s in self.native_scripts:
                scripts[script_hash(s)] = s

        for s in self._inputs_to_scripts.values():
            scripts[script_hash(s)] = s

        for s, _ in self._minting_script_to_redeemers:
            scripts[script_hash(s)] = s

        for s, _ in self._withdrawal_script_to_redeemers:
            scripts[script_hash(s)] = s

        for s, _ in self._certificate_script_to_redeemers:
            scripts[script_hash(s)] = s

        return list(scripts.values())

    @property
    def scripts(self) -> List[ScriptType]:
        scripts: Dict[ScriptHash, ScriptType] = {
            script_hash(s): s for s in self.all_scripts
        }
        s: ScriptType

        for s in self._reference_scripts:
            if script_hash(s) in scripts:
                scripts.pop(script_hash(s))

        return list(scripts.values())

    @property
    def datums(self) -> Dict[DatumHash, Datum]:
        return self._datums

    @property
    def _redeemer_list(self) -> List[Redeemer]:
        return (
            [r for r in self._inputs_to_redeemers.values() if r is not None]
            + [r for _, r in self._minting_script_to_redeemers if r is not None]
            + [r for _, r in self._withdrawal_script_to_redeemers if r is not None]
            + [r for _, r in self._certificate_script_to_redeemers if r is not None]
        )

    def redeemers(self) -> Redeemers:
        redeemer_list = self._redeemer_list

        # We have to serialize redeemers as a map if there are no redeemers
        if self.use_redeemer_map or not redeemer_list:
            redeemers = RedeemerMap()
            for r in redeemer_list:
                if r.tag is None:
                    raise InvalidArgumentException(
                        f"Redeemer tag is not set. Redeemer: {r}"
                    )
                if r.ex_units is None:
                    raise InvalidArgumentException(
                        f"Execution units are not set. Redeemer: {r}"
                    )
                k = RedeemerKey(r.tag, r.index)
                v = RedeemerValue(r.data, r.ex_units)
                redeemers[k] = v
            return redeemers
        else:
            return redeemer_list

    @property
    def script_data_hash(self) -> Optional[ScriptDataHash]:
        if self.datums or self._redeemer_list:
            cost_models = {}
            for s in self.all_scripts:
                version = -1
                if isinstance(s, PlutusScript):
                    version = s.version
                elif type(s) is bytes:
                    version = 1
                if version != -1:
                    cost_models[version - 1] = (
                        self.context.protocol_param.cost_models.get(
                            f"PlutusV{version}", {}
                        )
                    )
            return script_data_hash(
                self.redeemers(),
                NonEmptyOrderedSet(list(self.datums.values())),
                CostModels(cost_models),
            )
        else:
            return None

    def _calc_change(
        self, fees, inputs, outputs, address, precise_fee=False, respect_min_utxo=True
    ) -> List[TransactionOutput]:
        requested = Value(fees)
        for o in outputs:
            requested += o.amount

        provided = Value()
        for i in inputs:
            provided += i.output.amount

        if self.mint:
            provided.multi_asset += self.mint

        if self.withdrawals:
            for v in self.withdrawals.values():
                provided.coin += v

        provided.coin -= self._get_total_key_deposit()
        provided.coin -= self._get_total_proposal_deposit()
        provided.multi_asset.filter(
            lambda p, n, v: p in requested.multi_asset and n in requested.multi_asset[p]
        )
        if (
            provided.coin < requested.coin
            or requested.multi_asset > provided.multi_asset
        ):
            raise InvalidTransactionException(
                f"The input UTxOs cannot cover the transaction outputs and tx fee. \n"
                f"Inputs: {inputs} \n"
                f"Outputs: {outputs} \n"
                f"fee: {fees}"
            )

        change = provided - requested

        # Remove any asset that has 0 quantity
        if change.multi_asset:
            change.multi_asset = change.multi_asset.filter(lambda p, n, v: v > 0)

        change_output_arr = []

        # when there is only ADA left, simply use remaining coin value as change
        if not change.multi_asset:
            if respect_min_utxo and change.coin < min_lovelace_post_alonzo(
                TransactionOutput(address, change), self.context
            ):
                raise InsufficientUTxOBalanceException(
                    f"Not enough ADA left for change: {change.coin} but needs "
                    f"{min_lovelace_post_alonzo(TransactionOutput(address, change), self.context)}"
                )
            lovelace_change = Value(change.coin)
            change_output_arr.append(TransactionOutput(address, lovelace_change))

        # If there are multi asset in the change
        if change.multi_asset:
            # Split assets if size exceeds limits
            multi_asset_arr = self._pack_tokens_for_change(
                address, change, self.context.protocol_param.max_val_size
            )

            # Include minimum lovelace into each token output except for the last one
            for i, multi_asset in enumerate(multi_asset_arr):
                # Combine remainder of provided ADA with last MultiAsset for output
                # There may be rare cases where adding ADA causes size exceeds limit
                # We will revisit if it becomes an issue
                if respect_min_utxo and change.coin < min_lovelace_post_alonzo(
                    TransactionOutput(address, Value(0, multi_asset)), self.context
                ):
                    raise InsufficientUTxOBalanceException(
                        "Not enough ADA left to cover non-ADA assets in a change address"
                    )

                if i == len(multi_asset_arr) - 1:
                    # Include all ada in last output
                    change_value = Value(change.coin, multi_asset)
                else:
                    change_value = Value(0, multi_asset)
                    change_value.coin = min_lovelace_post_alonzo(
                        TransactionOutput(address, change_value), self.context
                    )

                change_output_arr.append(TransactionOutput(address, change_value))
                change -= change_value
                change.multi_asset = change.multi_asset.filter(lambda p, n, v: v > 0)

        return change_output_arr

    def _add_change_and_fee(
        self,
        change_address: Optional[Address],
        merge_change: Optional[bool] = False,
    ) -> TransactionBuilder:
        original_outputs = deepcopy(self.outputs)
        change_output_index = None

        def _merge_changes(changes):
            if change_output_index is not None and len(changes) == 1:
                # Add the leftover change to the TransactionOutput containing the change address
                self._outputs[change_output_index].amount = (
                    changes[0].amount + self._outputs[change_output_index].amount
                )
                # if we enforce that TransactionOutputs must use Values for `amount`, we can use += here

            else:
                self._outputs += changes

        if change_address:
            if merge_change:
                for idx, output in enumerate(original_outputs):
                    # Find any transaction outputs which already contain the change address
                    if change_address == output.address:
                        if change_output_index is None or output.lovelace == 0:
                            change_output_index = idx

            # Set fee to max
            self.fee = self._estimate_fee()

            changes = self._calc_change(
                self.fee,
                self.inputs,
                self.outputs,
                change_address,
                precise_fee=True,
                respect_min_utxo=not merge_change,
            )

            _merge_changes(changes)

        # With changes included, we can estimate the fee more precisely
        self.fee = self._estimate_fee()

        if change_address:
            self._outputs = original_outputs
            changes = self._calc_change(
                self.fee,
                self.inputs,
                self.outputs,
                change_address,
                precise_fee=True,
                respect_min_utxo=not merge_change,
            )

            _merge_changes(changes)

        return self

    def _adding_asset_make_output_overflow(
        self,
        output: TransactionOutput,
        current_assets: Asset,
        policy_id: ScriptHash,
        add_asset_name: AssetName,
        add_asset_val: int,
        max_val_size: int,
    ) -> bool:
        """Check if adding the asset will make output exceed maximum size limit

        Args:
            output (TransactionOutput): Current output
            current_assets (Asset): Current Assets to be included in output
            policy_id (ScriptHash): Policy id containing the MultiAsset
            add_asset_name (AssetName): Name of asset to add to current MultiAsset
            add_asset_val (int): Value of asset to add to current MultiAsset
            max_val_size (int): maximum size limit for output

        Returns:
            bool: whether adding asset will make output greater than maximum size limit
        """
        attempt_assets = deepcopy(current_assets)
        attempt_assets += Asset({add_asset_name: add_asset_val})
        attempt_multi_asset = MultiAsset({policy_id: attempt_assets})

        new_amount = Value(0, attempt_multi_asset)
        current_amount = deepcopy(output.amount)
        attempt_amount = new_amount + current_amount

        # Calculate minimum ada requirements for more precise value size
        required_lovelace = min_lovelace_post_alonzo(
            TransactionOutput(output.address, attempt_amount), self.context
        )
        attempt_amount.coin = required_lovelace

        return len(attempt_amount.to_cbor()) > max_val_size

    def _pack_tokens_for_change(
        self,
        change_address: Optional[Address],
        change_estimator: Value,
        max_val_size: int,
    ) -> List[MultiAsset]:
        multi_asset_arr = []
        base_coin = Value(coin=change_estimator.coin)
        change_address = change_address or Address(FAKE_VKEY.hash())
        output = TransactionOutput(change_address, base_coin)

        # iteratively add tokens to output
        for policy_id, assets in change_estimator.multi_asset.items():
            temp_multi_asset = MultiAsset()
            temp_value = Value(coin=0)
            temp_assets = Asset()
            old_amount = deepcopy(output.amount)
            for asset_name, asset_value in assets.items():
                if self._adding_asset_make_output_overflow(
                    output,
                    temp_assets,
                    policy_id,
                    asset_name,
                    asset_value,
                    max_val_size,
                ):
                    # Insert current assets as one group if current assets isn't null
                    # This handles edge case when first Asset from next policy will cause overflow
                    if temp_assets:
                        temp_multi_asset += MultiAsset({policy_id: temp_assets})
                        temp_value.multi_asset = temp_multi_asset
                        output.amount += temp_value
                    multi_asset_arr.append(output.amount.multi_asset)

                    # Create a new output
                    base_coin = Value(coin=0)
                    output = TransactionOutput(change_address, base_coin)

                    # Continue building output from where we stopped
                    old_amount = deepcopy(output.amount)
                    temp_multi_asset = MultiAsset()
                    temp_value = Value()
                    temp_assets = Asset()

                temp_assets += Asset({asset_name: asset_value})

            # Assess assets in buffer
            temp_multi_asset += MultiAsset({policy_id: temp_assets})
            temp_value.multi_asset = temp_multi_asset
            output.amount += temp_value

            # Calculate min lovelace required for more precise size
            updated_amount = deepcopy(output.amount)
            required_lovelace = min_lovelace_post_alonzo(
                TransactionOutput(change_address, updated_amount), self.context
            )
            updated_amount.coin = required_lovelace

            if len(updated_amount.to_cbor()) > max_val_size:
                output.amount = old_amount
                break

        multi_asset_arr.append(output.amount.multi_asset)
        # Remove records where MultiAsset is null due to overflow of adding
        # items at the beginning of next policy to previous policy MultiAssets
        return multi_asset_arr

    def _required_signer_vkey_hashes(self) -> Set[VerificationKeyHash]:
        return set(self.required_signers) if self.required_signers else set()

    def _input_vkey_hashes(self) -> Set[VerificationKeyHash]:
        results = set()
        for i in self.inputs + list(self.collaterals):
            if isinstance(i.output.address.payment_part, VerificationKeyHash):
                results.add(i.output.address.payment_part)
        return results

    def _certificate_vkey_hashes(self) -> Set[VerificationKeyHash]:
        results = set()

        def _check_and_add_vkey(stake_credential: StakeCredential):
            if isinstance(stake_credential.credential, VerificationKeyHash):
                results.add(stake_credential.credential)

        if self.certificates:
            for cert in self.certificates:
                if isinstance(
                    cert,
                    (
                        StakeRegistration,
                        StakeDeregistration,
                        StakeDelegation,
                        StakeRegistrationConway,
                        StakeDeregistrationConway,
                        VoteDelegation,
                        StakeAndVoteDelegation,
                        StakeRegistrationAndDelegation,
                        StakeRegistrationAndVoteDelegation,
                        StakeRegistrationAndDelegationAndVoteDelegation,
                    ),
                ):
                    _check_and_add_vkey(cert.stake_credential)
                elif isinstance(cert, RegDRepCert):
                    _check_and_add_vkey(cert.drep_credential)
                elif isinstance(cert, PoolRegistration):
                    results.add(cert.pool_params.operator)
                elif isinstance(cert, PoolRetirement):
                    results.add(cert.pool_keyhash)
        return results

    def _vote_vkey_hashes(self) -> Set[VerificationKeyHash]:
        results = set()

        if self.voting_procedures:
            for voter in self.voting_procedures:
                if isinstance(voter.credential, VerificationKeyHash):
                    results.add(voter.credential)
        return results

    def _get_total_key_deposit(self):
        stake_registration_certs = set()
        stake_registration_certs_with_explicit_deposit = set()
        stake_pool_registration_certs = set()

        protocol_params = self.context.protocol_param

        if self.certificates:
            for cert in self.certificates:
                if isinstance(cert, StakeRegistration):
                    stake_registration_certs.add(cert.stake_credential.credential)
                elif isinstance(
                    cert,
                    (
                        RegDRepCert,
                        StakeRegistrationConway,
                        StakeRegistrationAndDelegation,
                        StakeRegistrationAndVoteDelegation,
                        StakeRegistrationAndDelegationAndVoteDelegation,
                    ),
                ):
                    stake_registration_certs_with_explicit_deposit.add(cert.coin)
                elif (
                    isinstance(cert, PoolRegistration)
                    and self.initial_stake_pool_registration
                ):
                    stake_pool_registration_certs.add(cert.pool_params.operator)

        stake_registration_deposit = protocol_params.key_deposit * len(
            stake_registration_certs
        ) + sum(stake_registration_certs_with_explicit_deposit)
        stake_pool_registration_deposit = protocol_params.pool_deposit * len(
            stake_pool_registration_certs
        )
        return stake_registration_deposit + stake_pool_registration_deposit

    def _get_total_proposal_deposit(self):
        proposal_deposit = 0
        if self.proposal_procedures:
            for proposal in self.proposal_procedures:
                proposal_deposit += proposal.deposit
        return proposal_deposit

    def _withdrawal_vkey_hashes(self) -> Set[VerificationKeyHash]:
        results = set()

        if self.withdrawals:
            for k in self.withdrawals:
                address = Address.from_primitive(k)
                if address.address_type == AddressType.NONE_KEY:
                    results.add(address.staking_part)

        return results

    def _native_scripts_vkey_hashes(self) -> Set[VerificationKeyHash]:
        results = set()

        def _dfs(script: NativeScript):
            tmp = set()
            if isinstance(script, ScriptPubkey):
                tmp.add(script.key_hash)
            elif isinstance(script, (ScriptAll, ScriptAny)):
                for s in script.native_scripts:
                    tmp.update(_dfs(s))
            return tmp

        if self.native_scripts:
            for script in self.native_scripts:
                results.update(_dfs(script))

        return results

    def _set_redeemer_index(self):
        # Set redeemers' index according to section 4.1 in
        # https://hydra.iohk.io/build/13099856/download/1/alonzo-changes.pdf
        #
        # There is no way to determine certificate index here

        if self.mint:
            sorted_mint_policies = sorted(self.mint.keys(), key=lambda x: x.to_cbor())
        else:
            sorted_mint_policies = []
        if self.withdrawals:
            sorted_withdrawals = sorted(self.withdrawals.keys())
        else:
            sorted_withdrawals = []

        for i, utxo in enumerate(self.inputs):
            if (
                utxo in self._inputs_to_redeemers
                and self._inputs_to_redeemers[utxo].tag == RedeemerTag.SPEND
            ):
                self._inputs_to_redeemers[utxo].index = i

        for script, redeemer in self._minting_script_to_redeemers:
            if redeemer is not None:
                redeemer.index = sorted_mint_policies.index(script_hash(script))

        for script, redeemer in self._withdrawal_script_to_redeemers:
            if redeemer is not None:
                script_staking_credential = Address(
                    staking_part=script_hash(script), network=self.context.network
                )
                redeemer.index = sorted_withdrawals.index(
                    script_staking_credential.to_primitive()
                )

        self._redeemer_list.sort(key=lambda r: r.index)

    def _build_tx_body(self) -> TransactionBody:
        tx_body = TransactionBody(
            OrderedSet([i.input for i in self.inputs]),
            self.outputs,
            fee=self.fee,
            ttl=self.ttl,
            mint=self.mint,
            auxiliary_data_hash=(
                self.auxiliary_data.hash() if self.auxiliary_data else None
            ),
            script_data_hash=self.script_data_hash,
            required_signers=(
                NonEmptyOrderedSet(self.required_signers)
                if self.required_signers
                else None
            ),
            validity_start=self.validity_start,
            collateral=(
                NonEmptyOrderedSet([c.input for c in self.collaterals])
                if self.collaterals
                else None
            ),
            certificates=self.certificates,
            withdraws=self.withdrawals,
            collateral_return=self._collateral_return,
            total_collateral=self._total_collateral,
            reference_inputs=(
                NonEmptyOrderedSet(
                    [
                        i.input if isinstance(i, UTxO) else i
                        for i in self.reference_inputs
                    ]
                )
                if self.reference_inputs
                else None
            ),
            # Add new governance fields
            voting_procedures=(
                self.voting_procedures if self.voting_procedures else None
            ),
            proposal_procedures=(
                self.proposal_procedures if self.proposal_procedures else None
            ),
            current_treasury_value=(
                self.current_treasury_value if self.current_treasury_value else None
            ),
            donation=self.donation if self.donation else None,
        )
        return tx_body

    def _build_required_vkeys(self) -> Set[VerificationKeyHash]:
        vkey_hashes = self._input_vkey_hashes()
        vkey_hashes.update(self._required_signer_vkey_hashes())
        vkey_hashes.update(self._native_scripts_vkey_hashes())
        vkey_hashes.update(self._certificate_vkey_hashes())
        vkey_hashes.update(self._withdrawal_vkey_hashes())
        vkey_hashes.update(self._vote_vkey_hashes())
        return vkey_hashes

    def _witness_count(self) -> int:
        return self.witness_override or len(self._build_required_vkeys())

    def _build_fake_vkey_witnesses(self) -> NonEmptyOrderedSet[VerificationKeyWitness]:
        witnesses = []
        for i in range(self._witness_count()):
            # Convert index to 32 bytes and use AND operation to create unique keys
            i_bytes = i.to_bytes(32, "big")
            unique_vkey = VerificationKey.from_primitive(
                bytes(
                    x & y
                    for x, y in zip(
                        bytes.fromhex(
                            "5797dc2cc919dfec0bb849551ebdf30d96e5cbe0f33f734a87fe826db30f7ef9"
                        ),
                        i_bytes,
                    )
                )
            )
            unique_sig = bytes(
                x & y
                for x, y in zip(
                    bytes.fromhex(
                        "577ccb5b487b64e396b0976c6f71558e52e44ad254db7d06dfb79843e5441a5d"
                        "763dd42adcf5e8805d70373722ebbce62a58e3f30dd4560b9a898b8ceeab6a03"
                    ),
                    i_bytes + i_bytes,  # 64 bytes for signature
                )
            )
            witnesses.append(VerificationKeyWitness(unique_vkey, unique_sig))
        return NonEmptyOrderedSet(witnesses)

    def _build_fake_witness_set(self) -> TransactionWitnessSet:
        witness_set = self.build_witness_set()
        if self._witness_count() > 0:
            witness_set.vkey_witnesses = self._build_fake_vkey_witnesses()

        return witness_set

    def _build_full_fake_tx(self) -> Transaction:
        tx_body = self._build_tx_body()

        if tx_body.fee == 0:
            # When fee is not specified, we will use max possible fee to fill in the fee field.
            # This will make sure the size of fee field itself is taken into account during fee estimation.
            tx_body.fee = max_tx_fee(self.context)

        witness = self._build_fake_witness_set()
        tx = Transaction(tx_body, witness, True, self.auxiliary_data)
        size = len(tx.to_cbor())
        if size > self.context.protocol_param.max_tx_size:
            raise InvalidTransactionException(
                f"Transaction size ({size}) exceeds the max limit "
                f"({self.context.protocol_param.max_tx_size}). Please try reducing the "
                f"number of inputs or outputs."
            )

        return tx

    def build_witness_set(
        self, remove_dup_script: bool = False
    ) -> TransactionWitnessSet:
        """Build a transaction witness set, excluding verification key witnesses.
        This function is especially useful when the transaction involves Plutus scripts.

        Args:
            remove_dup_script (bool): Whether to remove scripts, that are already attached to inputs,
             from the witness set.

        Returns:
            TransactionWitnessSet: A transaction witness set without verification key witnesses.
        """

        native_scripts: NonEmptyOrderedSet[NativeScript] = NonEmptyOrderedSet()
        plutus_v1_scripts: NonEmptyOrderedSet[PlutusV1Script] = NonEmptyOrderedSet()
        plutus_v2_scripts: NonEmptyOrderedSet[PlutusV2Script] = NonEmptyOrderedSet()
        plutus_v3_scripts: NonEmptyOrderedSet[PlutusV3Script] = NonEmptyOrderedSet()
        plutus_data: NonEmptyOrderedSet[Any] = NonEmptyOrderedSet()

        input_scripts = (
            {
                script_hash(i.output.script)
                for i in self.inputs
                if i.output.script is not None
            }
            if remove_dup_script
            else {}
        )

        for datum in self.datums.values():
            plutus_data.append(datum)

        for script in self.scripts:
            if script_hash(script) not in input_scripts:
                if isinstance(script, NativeScript):
                    native_scripts.append(script)
                elif isinstance(script, PlutusV1Script):
                    plutus_v1_scripts.append(script)
                elif type(script) is bytes:
                    plutus_v1_scripts.append(PlutusV1Script(script))
                elif isinstance(script, PlutusV2Script):
                    plutus_v2_scripts.append(script)
                elif isinstance(script, PlutusV3Script):
                    plutus_v3_scripts.append(script)
                else:
                    raise InvalidArgumentException(
                        f"Unsupported script type: {type(script)}"
                    )

        witness_set = TransactionWitnessSet(
            native_scripts=native_scripts if native_scripts else None,
            plutus_v1_script=plutus_v1_scripts if plutus_v1_scripts else None,
            plutus_v2_script=plutus_v2_scripts if plutus_v2_scripts else None,
            plutus_v3_script=plutus_v3_scripts if plutus_v3_scripts else None,
            redeemer=self.redeemers() if self._redeemer_list else None,
            plutus_data=plutus_data if plutus_data else None,
        )
        witness_set.convert_to_latest_spec()
        return witness_set

    def _ensure_no_input_exclusion_conflict(self):
        intersection = set(self.inputs).intersection(set(self.excluded_inputs))
        if intersection:
            raise TransactionBuilderException(
                f"Found common UTxOs between UTxO inputs and UTxO excluded_inputs: "
                f"{intersection}."
            )

    def _ref_script_size(self):
        ref_script_size = 0
        for s in self._reference_scripts:
            if isinstance(s, NativeScript):
                ref_script_size += len(s.to_cbor())
            else:
                ref_script_size += len(s)
        return ref_script_size

    def _estimate_fee(self):
        plutus_execution_units = ExecutionUnits(0, 0)
        for redeemer in self._redeemer_list:
            plutus_execution_units += redeemer.ex_units

        estimated_fee = fee(
            self.context,
            len(self._build_full_fake_tx().to_cbor()),
            plutus_execution_units.steps,
            plutus_execution_units.mem,
            self._ref_script_size(),
        )
        if self.fee_buffer is not None:
            estimated_fee += self.fee_buffer

        return estimated_fee

    @log_state
    def build(
        self,
        change_address: Optional[Address] = None,
        merge_change: Optional[bool] = False,
        collateral_change_address: Optional[Address] = None,
        auto_validity_start_offset: Optional[int] = None,
        auto_ttl_offset: Optional[int] = None,
        auto_required_signers: Optional[bool] = None,
    ) -> TransactionBody:
        """Build a transaction body from all constraints set through the builder.

        Args:
            change_address (Optional[Address]): Address to which changes will be returned. If not provided, the
                transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
            merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
                will be directly added to that transaction output, instead of being added as a separate output.
            collateral_change_address (Optional[Address]): Address to which collateral changes will be returned.
            auto_validity_start_offset (Optional[int]): Automatically set the validity start interval of the transaction
                to the current slot number + the given offset (default -1000).
                A manually set validity start will always take precedence.
            auto_ttl_offset (Optional[int]): Automatically set the validity end interval (ttl) of the transaction
                to the current slot number + the given offset (default 10_000).
                A manually set ttl will always take precedence.
            auto_required_signers (Optional[bool]): Automatically add all pubkeyhashes of transaction inputs
                to required signatories (default only for Smart Contract transactions).
                Manually set required signers will always take precedence.

        Returns:
            TransactionBody: A transaction body.
        """
        self._ensure_no_input_exclusion_conflict()

        # only automatically set the validity interval and required signers if scripts are involved
        is_smart = bool(self.all_scripts)

        # Automatically set the validity range to a tight value around transaction creation
        if (
            is_smart or auto_validity_start_offset is not None
        ) and self.validity_start is None:
            last_slot = self.context.last_block_slot
            # If None is provided, the default value is -1000
            if auto_validity_start_offset is None:
                auto_validity_start_offset = -1000
            self.validity_start = max(0, last_slot + auto_validity_start_offset)

        if (is_smart or auto_ttl_offset is not None) and self.ttl is None:
            last_slot = self.context.last_block_slot
            # If None is provided, the default value is 10_000
            if auto_ttl_offset is None:
                auto_ttl_offset = 10_000
            self.ttl = max(0, last_slot + auto_ttl_offset)

        selected_utxos = []
        selected_amount = Value()
        for i in self.inputs:
            selected_utxos.append(i)
            selected_amount += i.output.amount

        if self.mint:
            # Add positive minted amounts to the selected amount (=source)
            for pid, m in self.mint.items():
                for tkn, am in m.items():
                    if am > 0:
                        selected_amount += Value(
                            multi_asset=MultiAsset({pid: Asset({tkn: am})})
                        )

        if self.withdrawals:
            for v in self.withdrawals.values():
                selected_amount.coin += v

        can_merge_change = False
        if merge_change:
            for o in self.outputs:
                if o.address == change_address:
                    can_merge_change = True
                    break

        selected_amount.coin -= self._get_total_key_deposit()
        selected_amount.coin -= self._get_total_proposal_deposit()

        requested_amount = Value()
        for o in self.outputs:
            requested_amount += o.amount

        if self.mint:
            # Add negative minted amounts to the requested amount (=sink)
            for pid, m in self.mint.items():
                for tkn, am in m.items():
                    if am < 0:
                        requested_amount += Value(
                            multi_asset=MultiAsset({pid: Asset({tkn: -am})})
                        )

        # Include min fees associated as part of requested amount
        requested_amount += self._estimate_fee()

        # Trim off assets that are not requested because they will be returned as changes eventually.
        trimmed_selected_amount = Value(
            selected_amount.coin,
            selected_amount.multi_asset.filter(
                lambda p, n, v: p in requested_amount.multi_asset
                and n in requested_amount.multi_asset[p]
            ),
        )

        unfulfilled_amount = requested_amount - trimmed_selected_amount

        remaining = trimmed_selected_amount - requested_amount
        remaining.multi_asset = remaining.multi_asset.filter(lambda p, n, v: v > 0)
        remaining.coin = max(0, remaining.coin)

        if change_address is not None and not can_merge_change:
            # If change address is provided and remainder is smaller than minimum ADA required in change,
            # we need to select additional UTxOs available from the address
            if unfulfilled_amount.coin < 0:

                unfulfilled_amount.coin = max(
                    0,
                    unfulfilled_amount.coin
                    + min_lovelace_post_alonzo(
                        TransactionOutput(
                            change_address, selected_amount - trimmed_selected_amount
                        ),
                        self.context,
                    ),
                )
        else:
            unfulfilled_amount.coin = max(0, unfulfilled_amount.coin)

        # Clean up all non-positive assets
        unfulfilled_amount.multi_asset = unfulfilled_amount.multi_asset.filter(
            lambda p, n, v: v > 0
        )

        # Create a set of all seen utxos in addition to other utxo lists.
        # We need this set to avoid adding the same utxo twice.
        # The reason of not turning all utxo lists into sets is that we want to keep the order of utxos and make
        # utxo selection deterministic.
        seen_utxos = set(selected_utxos)

        # When there are positive coin or native asset quantity in unfulfilled Value
        if Value() < unfulfilled_amount:
            additional_utxo_pool = []
            additional_amount = Value()

            for utxo in self.potential_inputs:
                additional_amount += utxo.output.amount
                seen_utxos.add(utxo)
                additional_utxo_pool.append(utxo)

            for address in self.input_addresses:
                for utxo in self.context.utxos(address):
                    if (
                        utxo not in seen_utxos
                        and utxo not in self.excluded_inputs
                        and utxo.output.script is None
                    ):
                        additional_utxo_pool.append(utxo)
                        additional_amount += utxo.output.amount
                        seen_utxos.add(utxo)

            for index, selector in enumerate(self.utxo_selectors):
                try:
                    selected, _ = selector.select(
                        additional_utxo_pool,
                        [
                            TransactionOutput(
                                Address(FAKE_VKEY.hash()), unfulfilled_amount
                            )
                        ],
                        self.context,
                        include_max_fee=False,
                        respect_min_utxo=not can_merge_change,
                        existing_amount=remaining,
                    )

                    for s in selected:
                        selected_amount += s.output.amount
                        selected_utxos.append(s)
                    break

                except UTxOSelectionException as e:
                    if index < len(self.utxo_selectors) - 1:
                        logger.info(e)
                        logger.info(f"{selector} failed. Trying next selector.")
                    else:
                        trimmed_additional_amount = Value(
                            additional_amount.coin,
                            additional_amount.multi_asset.filter(
                                lambda p, n, v: p in requested_amount.multi_asset
                                and n in requested_amount.multi_asset[p]
                            ),
                        )
                        diff = (
                            requested_amount
                            - trimmed_selected_amount
                            - trimmed_additional_amount
                        )
                        diff.multi_asset = diff.multi_asset.filter(
                            lambda p, n, v: v > 0
                        )
                        raise UTxOSelectionException(
                            f"All UTxO selectors failed.\n"
                            f"Requested output:\n {requested_amount} \n"
                            f"Pre-selected inputs:\n {selected_amount} \n"
                            f"Additional UTxO pool:\n {additional_utxo_pool} \n"
                            f"Unfulfilled amount:\n {diff}"
                        )

        selected_utxos.sort(
            key=lambda utxo: (str(utxo.input.transaction_id), utxo.input.index)
        )

        self.inputs[:] = selected_utxos[:]

        # Automatically set the required signers for smart transactions
        if (
            is_smart and auto_required_signers is not False
        ) and self.required_signers is None:
            # Collect all signatories from explicitly defined
            # transaction inputs and collateral inputs, and input addresses
            self.required_signers = list(self._input_vkey_hashes())

        self._set_redeemer_index()

        self._set_collateral_return(collateral_change_address or change_address)

        self._update_execution_units(
            change_address, merge_change, collateral_change_address
        )

        self._add_change_and_fee(change_address, merge_change=merge_change)

        tx_body = self._build_tx_body()

        return tx_body

    def _should_add_collateral_return(self, collateral_return: Value) -> bool:
        """Check if it is necessary to add a collateral return output.

        Args:
            collateral_return (Value): The potential collateral return amount.

        Returns:
            bool: True if a collateral return output should be added, False otherwise.
        """
        return (
            collateral_return.coin > max(self.collateral_return_threshold, 1_000_000)
            or collateral_return.multi_asset.count(lambda p, n, v: v > 0) > 0
        )

    def _set_collateral_return(self, collateral_return_address: Optional[Address]):
        """Calculate and set the change returned from the collateral inputs.

        Args:
            collateral_return_address (Address): Address to which the collateral change will be returned.
        """
        witnesses = self._build_fake_witness_set()

        # Make sure there is at least one script input
        if (
            not witnesses.plutus_v1_script
            and not witnesses.plutus_v2_script
            and not witnesses.plutus_v3_script
            and not self._reference_scripts
        ):
            return

        if not collateral_return_address:
            return

        collateral_amount = (
            max_tx_fee(context=self.context, ref_script_size=self._ref_script_size())
            * self.context.protocol_param.collateral_percent
            // 100
        )

        if not self.collaterals:
            tmp_val = Value()

            def _add_collateral_input(cur_total, candidate_inputs):
                cur_collateral_return = cur_total - collateral_amount

                while (
                    cur_total.coin < collateral_amount
                    or self._should_add_collateral_return(cur_collateral_return)
                    and 0
                    <= cur_collateral_return.coin
                    < min_lovelace_post_alonzo(
                        TransactionOutput(
                            collateral_return_address, cur_collateral_return
                        ),
                        self.context,
                    )
                ) and candidate_inputs:
                    candidate = candidate_inputs.pop()
                    if (
                        not candidate.output.address.address_type.name.startswith(
                            "SCRIPT"
                        )
                        and candidate.output.amount.coin > 2000000
                        and candidate not in self.collaterals
                    ):
                        self.collaterals.append(candidate)
                        cur_total += candidate.output.amount
                        cur_collateral_return = cur_total - collateral_amount

            sorted_inputs = sorted(
                self.inputs.copy(),
                key=lambda i: (len(i.output.to_cbor_hex()), -i.output.amount.coin),
            )
            _add_collateral_input(tmp_val, sorted_inputs)

            if tmp_val.coin < collateral_amount:
                sorted_inputs = sorted(
                    self.potential_inputs,
                    key=lambda i: (len(i.output.to_cbor_hex()), -i.output.amount.coin),
                )
                _add_collateral_input(tmp_val, sorted_inputs)

            if tmp_val.coin < collateral_amount:
                sorted_inputs = sorted(
                    self.context.utxos(collateral_return_address),
                    key=lambda i: (len(i.output.to_cbor_hex()), -i.output.amount.coin),
                )
                _add_collateral_input(tmp_val, sorted_inputs)

        total_input = Value()

        for utxo in self.collaterals:
            total_input += utxo.output.amount

        if collateral_amount > total_input.coin:
            raise ValueError(
                f"Minimum collateral amount {collateral_amount} is greater than total "
                f"provided collateral inputs {total_input}"
            )
        else:
            return_amount = total_input - collateral_amount

            if not self._should_add_collateral_return(return_amount):
                return  # No need to return collateral if the remaining amount is too small

            min_lovelace_val = min_lovelace_post_alonzo(
                TransactionOutput(collateral_return_address, return_amount),
                self.context,
            )
            if min_lovelace_val > return_amount.coin:
                raise ValueError(
                    f"Minimum lovelace amount for collateral return {min_lovelace_val} is "
                    f"greater than collateral change {return_amount.coin}. Please provide more collateral inputs."
                )
            else:
                self._collateral_return = TransactionOutput(
                    collateral_return_address, total_input - collateral_amount
                )
                self._total_collateral = collateral_amount

    def _update_execution_units(
        self,
        change_address: Optional[Address] = None,
        merge_change: Optional[bool] = False,
        collateral_change_address: Optional[Address] = None,
    ):
        if self._should_estimate_execution_units:
            estimated_execution_units = self._estimate_execution_units(
                change_address, merge_change, collateral_change_address
            )
            for r in self._redeemer_list:
                assert (
                    r.tag is not None
                ), "Expected tag of redeemer to be set, but found None"
                tagname = r.tag.name.lower()
                key = f"{tagname}:{r.index}"
                if (
                    key not in estimated_execution_units
                    or estimated_execution_units[key] is None
                ):
                    raise TransactionBuilderException(
                        f"Cannot find execution unit for redeemer: {r} "
                        f"in estimated execution units: {estimated_execution_units}"
                    )
                r.ex_units = estimated_execution_units[key]
                r.ex_units.mem = int(
                    r.ex_units.mem * (1 + self.execution_memory_buffer)
                )
                r.ex_units.steps = int(
                    r.ex_units.steps * (1 + self.execution_step_buffer)
                )

    def _estimate_execution_units(
        self,
        change_address: Optional[Address] = None,
        merge_change: Optional[bool] = False,
        collateral_change_address: Optional[Address] = None,
    ) -> Dict[str, ExecutionUnits]:
        # Create a deep copy of current builder, so we won't mess up current builder's internal states
        tmp_builder = TransactionBuilder(self.context)
        for f in fields(self):
            if f.name not in ("context",):
                setattr(tmp_builder, f.name, deepcopy(getattr(self, f.name)))
        tmp_builder._should_estimate_execution_units = False
        self._should_estimate_execution_units = False
        tx_body = tmp_builder.build(
            change_address, merge_change, collateral_change_address
        )
        witness_set = tmp_builder._build_fake_witness_set()
        tx = Transaction(
            tx_body, witness_set, auxiliary_data=tmp_builder.auxiliary_data
        )

        return self.context.evaluate_tx(tx)

    def build_and_sign(
        self,
        signing_keys: List[Union[SigningKey, ExtendedSigningKey]],
        change_address: Optional[Address] = None,
        merge_change: Optional[bool] = False,
        collateral_change_address: Optional[Address] = None,
        auto_validity_start_offset: Optional[int] = None,
        auto_ttl_offset: Optional[int] = None,
        auto_required_signers: Optional[bool] = None,
        force_skeys: Optional[bool] = False,
    ) -> Transaction:
        """Build a transaction body from all constraints set through the builder and sign the transaction with
        provided signing keys.

        Args:
            signing_keys (List[Union[SigningKey, ExtendedSigningKey]]): A list of signing keys that will be used to
                sign the transaction.
            change_address (Optional[Address]): Address to which changes will be returned. If not provided, the
                transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
            merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
                will be directly added to that transaction output, instead of being added as a separate output.
            collateral_change_address (Optional[Address]): Address to which collateral changes will be returned.
            auto_validity_start_offset (Optional[int]): Automatically set the validity start interval of the transaction
                to the current slot number + the given offset (default -1000).
                A manually set validity start will always take precedence.
            auto_ttl_offset (Optional[int]): Automatically set the validity end interval (ttl) of the transaction
                to the current slot number + the given offset (default 10_000).
                A manually set ttl will always take precedence.
            auto_required_signers (Optional[bool]): Automatically add all pubkeyhashes of transaction inputs
                and the given signers to required signatories (default only for Smart Contract transactions).
                Manually set required signers will always take precedence.
            force_skeys (Optional[bool]): Whether to force the use of signing keys for signing the transaction.
                Default is False, which means that provided signing keys will only be used to sign the transaction if
                they are actually required by the transaction. This is useful to reduce tx fees by not including
                unnecessary signatures. If set to True, all provided signing keys will be used to sign the transaction.

        Returns:
            Transaction: A signed transaction.
        """
        # The given signers should be required signers if they weren't added yet
        if auto_required_signers and self.scripts and not self.required_signers:
            # Collect all signatories from explicitly defined
            # transaction inputs and collateral inputs, and input addresses
            self.required_signers = [
                s.to_verification_key().hash() for s in signing_keys
            ]

        tx_body = self.build(
            change_address=change_address,
            merge_change=merge_change,
            collateral_change_address=collateral_change_address,
            auto_validity_start_offset=auto_validity_start_offset,
            auto_ttl_offset=auto_ttl_offset,
            auto_required_signers=auto_required_signers,
        )
        witness_set = self.build_witness_set(True)
        witness_set.vkey_witnesses = NonEmptyOrderedSet()

        required_vkeys = self._build_required_vkeys()

        for signing_key in set(signing_keys):
            vkey_hash = signing_key.to_verification_key().hash()
            if not force_skeys and vkey_hash not in required_vkeys:
                logger.warning(
                    f"Verification key hash {vkey_hash} is not required for this tx."
                )
                continue
            signature = signing_key.sign(tx_body.hash())
            witness_set.vkey_witnesses.append(
                VerificationKeyWitness(signing_key.to_verification_key(), signature)
            )

        if len(witness_set.vkey_witnesses) == 0:
            witness_set.vkey_witnesses = None

        return Transaction(tx_body, witness_set, auxiliary_data=self.auxiliary_data)

    # Add helper methods for governance operations
    def add_vote(
        self,
        voter: Voter,
        gov_action_id: GovActionId,
        vote: Vote,
        anchor: Optional[Anchor] = None,
    ) -> TransactionBuilder:
        """Add a vote to the transaction.

        Args:
            voter: The voter casting the vote
            gov_action_id: The ID of the governance action being voted on
            vote: The vote being cast (YES/NO/ABSTAIN)
            anchor: Optional metadata about the vote

        Returns:
            self: The transaction builder instance
        """
        if self.voting_procedures is None:
            self.voting_procedures = VotingProcedures()

        # Initialize the inner map if this is the first vote for this voter
        if voter not in self.voting_procedures:
            self.voting_procedures[voter] = GovActionIdToVotingProcedure()

        # Add the voting procedure for this specific governance action
        self.voting_procedures[voter][gov_action_id] = VotingProcedure(vote, anchor)

        return self

    def add_proposal(
        self,
        deposit: int,
        reward_account: bytes,
        gov_action: GovAction,
        anchor: Anchor,
    ) -> TransactionBuilder:
        """Add a governance proposal to the transaction.

        Args:
            deposit: The deposit amount required for the proposal
            reward_account: The reward account for the proposal
            gov_action: The governance action being proposed
            anchor: Metadata about the proposal

        Returns:
            self: The transaction builder instance
        """
        if self.proposal_procedures is None:
            self.proposal_procedures = NonEmptyOrderedSet()

        self.proposal_procedures.append(
            ProposalProcedure(
                deposit=deposit,
                reward_account=reward_account,
                gov_action=gov_action,
                anchor=anchor,
            )
        )
        return self

    def add_treasury_donation(self, amount: int) -> TransactionBuilder:
        """Add a donation to the treasury.

        Args:
            amount: The amount to donate (must be positive)

        Returns:
            self: The transaction builder instance
        """
        if amount <= 0:
            raise ValueError("Treasury donation amount must be positive")
        self.donation = amount
        return self



"""Cryptographic keys that could be used in address generation and transaction signing."""

from __future__ import annotations

import json
from typing import Optional, Type

from nacl.encoding import RawEncoder
from nacl.hash import blake2b
from nacl.public import PrivateKey
from nacl.signing import SigningKey as NACLSigningKey

from pycardano.crypto.bip32 import BIP32ED25519PrivateKey, HDWallet
from pycardano.exception import InvalidKeyTypeException
from pycardano.hash import VERIFICATION_KEY_HASH_SIZE, VerificationKeyHash
from pycardano.serialization import CBORSerializable, limit_primitive_type

__all__ = [
    "Key",
    "ExtendedSigningKey",
    "ExtendedVerificationKey",
    "VerificationKey",
    "SigningKey",
    "PaymentExtendedSigningKey",
    "PaymentExtendedVerificationKey",
    "PaymentSigningKey",
    "PaymentVerificationKey",
    "PaymentKeyPair",
    "StakeExtendedSigningKey",
    "StakeExtendedVerificationKey",
    "StakeSigningKey",
    "StakeVerificationKey",
    "StakeKeyPair",
    "StakePoolSigningKey",
    "StakePoolVerificationKey",
    "StakePoolKeyPair",
]


class Key(CBORSerializable):
    """A class that holds a cryptographic key and some metadata. e.g. signing key, verification key."""

    KEY_TYPE = ""
    DESCRIPTION = ""

    def __init__(
        self,
        payload: bytes,
        key_type: Optional[str] = None,
        description: Optional[str] = None,
    ):
        self._payload = payload
        self._key_type = key_type or self.KEY_TYPE
        self._description = description or self.KEY_TYPE

    @property
    def payload(self) -> bytes:
        return self._payload

    @property
    def key_type(self) -> str:
        return self._key_type

    @property
    def description(self) -> str:
        return self._description

    def to_primitive(self) -> bytes:
        return self.payload

    @classmethod
    @limit_primitive_type(bytes)
    def from_primitive(cls: Type["Key"], value: bytes) -> Key:
        return cls(value)

    def to_json(self, **kwargs) -> str:  # type: ignore
        """Serialize the key to JSON.

        The json output has three fields: "type", "description", and "cborHex".

        Returns:
            str: JSON representation of the key.
        """
        return json.dumps(
            {
                "type": self.key_type,
                "description": self.description,
                "cborHex": self.to_cbor_hex(),
            }
        )

    @classmethod
    def from_json(cls: Type[Key], data: str, validate_type=False) -> Key:  # type: ignore
        """Restore a key from a JSON string.

        Args:
            data (str): JSON string.
            validate_type (bool): Checks whether the type specified in json object is the same
                as the class's default type.

        Returns:
            Key: The key restored from JSON.

        Raises:
            InvalidKeyTypeException: When `validate_type=True` and the type in json is not equal to the default type
                of the Key class used.
        """
        obj = json.loads(data)

        if validate_type and obj["type"] != cls.KEY_TYPE:
            raise InvalidKeyTypeException(
                f"Expect key type: {cls.KEY_TYPE}, got {obj['type']} instead."
            )

        k = cls.from_cbor(obj["cborHex"])

        assert isinstance(k, cls)

        return cls(
            k.payload,
            key_type=obj["type"],
            description=obj["description"],
        )

    def __bytes__(self):
        return self.payload

    def __eq__(self, other):
        if not isinstance(other, Key):
            return False
        else:
            return (
                self.payload == other.payload
                and self.description == other.description
                and self.key_type == other.key_type
            )

    def __repr__(self) -> str:
        return self.to_json()

    def __hash__(self):
        return hash(self.payload)


class SigningKey(Key):
    def sign(self, data: bytes) -> bytes:
        signed_message = NACLSigningKey(self.payload).sign(data)
        return signed_message.signature

    def to_verification_key(self) -> VerificationKey:
        verification_key = NACLSigningKey(self.payload).verify_key
        return VerificationKey(
            bytes(verification_key),
            self.key_type.replace("Signing", "Verification"),
            self.description.replace("Signing", "Verification"),
        )

    @classmethod
    def generate(cls) -> SigningKey:
        signing_key = PrivateKey.generate()
        return cls(bytes(signing_key))


class VerificationKey(Key):
    def hash(self) -> VerificationKeyHash:
        """Compute a blake2b hash from the key

        Returns:
            VerificationKeyHash: Hash output in bytes.
        """
        return VerificationKeyHash(
            blake2b(self.payload, VERIFICATION_KEY_HASH_SIZE, encoder=RawEncoder)
        )

    @classmethod
    def from_signing_key(cls, key: SigningKey) -> VerificationKey:
        return key.to_verification_key()


class ExtendedSigningKey(Key):
    def sign(self, data: bytes) -> bytes:
        private_key = BIP32ED25519PrivateKey(self.payload[:64], self.payload[96:])
        return private_key.sign(data)

    def to_verification_key(self) -> ExtendedVerificationKey:
        return ExtendedVerificationKey(
            self.payload[64:],
            self.key_type.replace("Signing", "Verification"),
            self.description.replace("Signing", "Verification"),
        )

    @classmethod
    def from_hdwallet(cls, hdwallet: HDWallet) -> ExtendedSigningKey:
        if hdwallet.xprivate_key is None or hdwallet.chain_code is None:
            raise InvalidKeyTypeException(
                "The hdwallet doesn't contain extended private key or chain code info."
            )

        return cls(
            payload=hdwallet.xprivate_key + hdwallet.public_key + hdwallet.chain_code,
            key_type=cls.KEY_TYPE,
            description=cls.DESCRIPTION,
        )


class ExtendedVerificationKey(Key):
    def hash(self) -> VerificationKeyHash:
        """Compute a blake2b hash from the key, excluding chain code

        Returns:
            VerificationKeyHash: Hash output in bytes.
        """
        return self.to_non_extended().hash()

    @classmethod
    def from_signing_key(cls, key: ExtendedSigningKey) -> ExtendedVerificationKey:
        return key.to_verification_key()

    def to_non_extended(self) -> VerificationKey:
        """Get the 32-byte verification with chain code trimmed off

        Returns:
            VerificationKey: 32-byte verification with chain code trimmed off
        """
        return VerificationKey(self.payload[:32])


class PaymentSigningKey(SigningKey):
    KEY_TYPE = "PaymentSigningKeyShelley_ed25519"
    DESCRIPTION = "Payment Signing Key"


class PaymentVerificationKey(VerificationKey):
    KEY_TYPE = "PaymentVerificationKeyShelley_ed25519"
    DESCRIPTION = "Payment Verification Key"


class PaymentExtendedSigningKey(ExtendedSigningKey):
    KEY_TYPE = "PaymentExtendedSigningKeyShelley_ed25519_bip32"
    DESCRIPTION = "Payment Signing Key"


class PaymentExtendedVerificationKey(ExtendedVerificationKey):
    KEY_TYPE = "PaymentExtendedVerificationKeyShelley_ed25519_bip32"
    DESCRIPTION = "Payment Verification Key"


class PaymentKeyPair:
    def __init__(self, signing_key: SigningKey, verification_key: VerificationKey):
        self.signing_key = signing_key
        self.verification_key = verification_key

    @classmethod
    def generate(cls: Type[PaymentKeyPair]) -> PaymentKeyPair:
        signing_key = PaymentSigningKey.generate()
        return cls.from_signing_key(signing_key)

    @classmethod
    def from_signing_key(
        cls: Type[PaymentKeyPair], signing_key: SigningKey
    ) -> PaymentKeyPair:
        return cls(signing_key, PaymentVerificationKey.from_signing_key(signing_key))

    def __eq__(self, other):
        if isinstance(other, PaymentKeyPair):
            return (
                other.signing_key == self.signing_key
                and other.verification_key == self.verification_key
            )


class StakeSigningKey(SigningKey):
    KEY_TYPE = "StakeSigningKeyShelley_ed25519"
    DESCRIPTION = "Stake Signing Key"


class StakeVerificationKey(VerificationKey):
    KEY_TYPE = "StakeVerificationKeyShelley_ed25519"
    DESCRIPTION = "Stake Verification Key"


class StakeExtendedSigningKey(ExtendedSigningKey):
    KEY_TYPE = "StakeExtendedSigningKeyShelley_ed25519_bip32"
    DESCRIPTION = "Stake Signing Key"


class StakeExtendedVerificationKey(ExtendedVerificationKey):
    KEY_TYPE = "StakeExtendedVerificationKeyShelley_ed25519_bip32"
    DESCRIPTION = "Stake Verification Key"


class StakeKeyPair:
    def __init__(self, signing_key: SigningKey, verification_key: VerificationKey):
        self.signing_key = signing_key
        self.verification_key = verification_key

    @classmethod
    def generate(cls: Type[StakeKeyPair]) -> StakeKeyPair:
        signing_key = StakeSigningKey.generate()
        return cls.from_signing_key(signing_key)

    @classmethod
    def from_signing_key(
        cls: Type[StakeKeyPair], signing_key: SigningKey
    ) -> StakeKeyPair:
        return cls(signing_key, StakeVerificationKey.from_signing_key(signing_key))

    def __eq__(self, other):
        if isinstance(other, StakeKeyPair):
            return (
                other.signing_key == self.signing_key
                and other.verification_key == self.verification_key
            )


class StakePoolSigningKey(SigningKey):
    KEY_TYPE = "StakePoolSigningKey_ed25519"
    DESCRIPTION = "Stake Pool Operator Signing Key"


class StakePoolVerificationKey(VerificationKey):
    KEY_TYPE = "StakePoolVerificationKey_ed25519"
    DESCRIPTION = "Stake Pool Operator Verification Key"


class StakePoolKeyPair:
    def __init__(self, signing_key: SigningKey, verification_key: VerificationKey):
        self.signing_key = signing_key
        self.verification_key = verification_key

    @classmethod
    def generate(cls: Type[StakePoolKeyPair]) -> StakePoolKeyPair:
        signing_key = StakePoolSigningKey.generate()
        return cls.from_signing_key(signing_key)

    @classmethod
    def from_signing_key(
        cls: Type[StakePoolKeyPair], signing_key: SigningKey
    ) -> StakePoolKeyPair:
        return cls(signing_key, StakePoolVerificationKey.from_signing_key(signing_key))

    def __eq__(self, other):
        if isinstance(other, StakePoolKeyPair):
            return (
                other.signing_key == self.signing_key
                and other.verification_key == self.verification_key
            )



"""All type of hashes in Cardano ledger spec."""

from typing import Type, TypeVar, Union

from pycardano.serialization import CBORSerializable, limit_primitive_type

__all__ = [
    "VERIFICATION_KEY_HASH_SIZE",
    "SCRIPT_HASH_SIZE",
    "TRANSACTION_HASH_SIZE",
    "DATUM_HASH_SIZE",
    "AUXILIARY_DATA_HASH_SIZE",
    "POOL_KEY_HASH_SIZE",
    "SCRIPT_DATA_HASH_SIZE",
    "VRF_KEY_HASH_SIZE",
    "POOL_METADATA_HASH_SIZE",
    "REWARD_ACCOUNT_HASH_SIZE",
    "CIP129_PAYLOAD_SIZE",
    "ConstrainedBytes",
    "VerificationKeyHash",
    "ScriptHash",
    "PolicyHash",
    "PolicyId",
    "ScriptDataHash",
    "TransactionId",
    "DatumHash",
    "AuxiliaryDataHash",
    "PoolKeyHash",
    "PoolMetadataHash",
    "VrfKeyHash",
    "RewardAccountHash",
    "AnchorDataHash",
]

VERIFICATION_KEY_HASH_SIZE = 28
SCRIPT_HASH_SIZE = 28
SCRIPT_DATA_HASH_SIZE = 32
TRANSACTION_HASH_SIZE = 32
DATUM_HASH_SIZE = 32
AUXILIARY_DATA_HASH_SIZE = 32
POOL_KEY_HASH_SIZE = 28
POOL_METADATA_HASH_SIZE = 32
VRF_KEY_HASH_SIZE = 32
REWARD_ACCOUNT_HASH_SIZE = 29
ANCHOR_DATA_HASH_SIZE = 32
CIP129_PAYLOAD_SIZE = 29


T = TypeVar("T", bound="ConstrainedBytes")


class ConstrainedBytes(CBORSerializable):
    """A wrapped class of bytes with constrained size.

    Args:
        payload (bytes): Hash in bytes.
    """

    __slots__ = "_payload"

    MAX_SIZE = 32
    MIN_SIZE = 0

    def __init__(self, payload: bytes):
        assert self.MIN_SIZE <= len(payload) <= self.MAX_SIZE, (
            f"Invalid byte size: {len(payload)} for class {self.__class__}, "
            f"expected size range: [{self.MIN_SIZE}, {self.MAX_SIZE}]"
        )
        self._payload = payload

    def __bytes__(self):
        return self.payload

    def __hash__(self):
        return hash(self.payload)

    @property
    def payload(self) -> bytes:
        return self._payload

    def to_primitive(self) -> bytes:
        return self.payload

    @classmethod
    @limit_primitive_type(bytes, str)
    def from_primitive(cls: Type[T], value: Union[bytes, str]) -> T:
        if isinstance(value, str):
            value = bytes.fromhex(value)
        return cls(value)

    def __eq__(self, other):
        if isinstance(other, ConstrainedBytes):
            return self.payload == other.payload
        else:
            return False

    def __repr__(self):
        return f"{self.__class__.__name__}(hex='{self.payload.hex()}')"

    def __str__(self):
        return self.payload.hex()


class VerificationKeyHash(ConstrainedBytes):
    """Hash of a Cardano verification key."""

    MAX_SIZE = MIN_SIZE = VERIFICATION_KEY_HASH_SIZE


class ScriptHash(ConstrainedBytes):
    """Hash of a policy/plutus script."""

    MAX_SIZE = MIN_SIZE = SCRIPT_HASH_SIZE


class PolicyHash(ScriptHash):
    pass


class PolicyId(ScriptHash):
    pass


class ScriptDataHash(ConstrainedBytes):
    """Hash of script data. See
    https://github.com/input-output-hk/cardano-ledger/blob/525844be05adae151e82069dcd0000f3301ca0d0/eras/alonzo/
    test-suite/cddl-files/alonzo.cddl#L79-L86"""

    MAX_SIZE = MIN_SIZE = SCRIPT_DATA_HASH_SIZE


class TransactionId(ConstrainedBytes):
    """Hash of a transaction."""

    MAX_SIZE = MIN_SIZE = TRANSACTION_HASH_SIZE


class DatumHash(ConstrainedBytes):
    """Hash of a datum"""

    MAX_SIZE = MIN_SIZE = DATUM_HASH_SIZE


class AuxiliaryDataHash(ConstrainedBytes):
    """Hash of auxiliary data"""

    MAX_SIZE = MIN_SIZE = AUXILIARY_DATA_HASH_SIZE


class PoolKeyHash(VerificationKeyHash):
    """Hash of a stake pool"""

    MAX_SIZE = MIN_SIZE = POOL_KEY_HASH_SIZE


class PoolMetadataHash(ConstrainedBytes):
    """Hash of a stake pool metadata"""

    MAX_SIZE = MIN_SIZE = POOL_METADATA_HASH_SIZE


class VrfKeyHash(ConstrainedBytes):
    """Hash of a Cardano VRF key."""

    MAX_SIZE = MIN_SIZE = VRF_KEY_HASH_SIZE


class RewardAccountHash(ConstrainedBytes):
    """Hash of a Cardano VRF key."""

    MAX_SIZE = MIN_SIZE = REWARD_ACCOUNT_HASH_SIZE


class AnchorDataHash(ConstrainedBytes):
    """Hash of anchor data."""

    MAX_SIZE = MIN_SIZE = ANCHOR_DATA_HASH_SIZE
